"""A simple dictionary mapping resqml class names to more readable names."""

class_dict = {
    'obj_LocalDepth3dCrs': 'Coordinate Reference System (z is length)',
    'obj_LocalTime3dCrs': 'Coordinate Reference System (z is time)',
    'obj_IjkGridRepresentation': 'Grid (IJK)',
    'obj_UnstructuredColumnLayerGridRepresentation': 'Grid (Column Layer)',
    'obj_UnstructuredGridRepresentation': 'Grid (Unstructured)',
    'obj_GridConnectionSetRepresentation': 'Grid Connection Set',
    'obj_TruncatedIjkGridRepresentation': 'Grid (Truncated IJK)',
    'obj_TruncatedUnstructuredColumnLayerGridRepresentation': 'Grid (Truncated Column Layer)',
    'obj_GpGridRepresentation': 'Grid (General Purpose)',
    'obj_LocalGridSet': 'Local Grid Set',
    'obj_EpcExternalPartReference': 'HDF5 Reference',
    'obj_PropertyKind': 'Kind of Property',
    'obj_CategoricalProperty': 'Property (Categorical)',
    'obj_DiscreteProperty': 'Property (Discrete)',
    'obj_ContinuousProperty': 'Property (Continuous)',
    'obj_CategoricalPropertySeries': 'Property Series (Categorical)',
    'obj_DiscretePropertySeries': 'Property Series (Discrete)',
    'obj_ContinuousPropertySeries': 'Property Series (Continuous)',
    'obj_PropertySet': 'Property Set',
    'obj_TimeSeries': 'Time Series',
    'obj_WellboreTrajectoryRepresentation': 'Wellbore Trajectory',
    'obj_BlockedWellboreRepresentation': 'Blocked Wellbore',
    'obj_WellboreInterpretation': 'Wellbore Interpretation',
    'obj_WellboreFeature': 'Wellbore Feature',
    'obj_WellboreFrameRepresentation': 'Wellbore Frame',
    'obj_WellboreMarkerFrameRepresentation': 'Wellbore Marker Frame',
    'obj_DeviationSurveyRepresentation': 'Deviation Survey',
    'obj_MdDatum': 'Measured Depth Datum',
    # geo classes
    'obj_TectonicBoundaryFeature': 'Tectonic Boundary Feature (fault or fracture)',
    'obj_Grid2dRepresentation': 'Mesh (Grid 2D)',
    'obj_Grid2dSetRepresentation': 'Mesh Set (Grid 2D set)',
    'obj_SealedSurfaceFrameworkRepresentation': 'Sealed Surface Framework',
    'obj_SealedVolumeFrameworkRepresentation': 'Sealed Volume Framework',
    'obj_NonSealedSurfaceFrameworkRepresentation': 'Non-sealed Surface Framework',
    'obj_SeismicLineSetFeature': 'Seismic Line Set Feature',
    'obj_SeismicLineFeature': 'Seismic Line Feature',
    'obj_SeismicLatticeFeature': 'Seismic Lattice Feature',
    'obj_GeneticBoundaryFeature': 'Genetic Boundary Feature (horizon or geobody)',
    'obj_FaultInterpretation': 'Fault Interpretation',
    'obj_HorizonInterpretation': 'Horizon Interpretation',
    'obj_GenericFeatureInterpretation': 'Generic Feature Interpretation',
    'obj_EarthModelInterpretation': 'Earth Model Interpretation',
    'obj_OrganizationFeature': 'Organization Feature',
    'obj_StructuralOrganizationInterpretation': 'Structural Organization Interpretation',
    'obj_BoundaryFeature': 'Boundary Feature',
    'obj_BoundaryFeatureInterpretation': 'Boundary Feature Interpretation',
    'obj_FluidBoundaryFeature': 'Fluid Boundary Feature',
    'obj_FrontierFeature': 'Frontier Feature (area of interest boundary)',
    'obj_GeobodyFeature': 'Geobody Feature',
    'obj_GeobodyInterpretation': 'Geobody Interpretation',
    'obj_GeobodyBoundaryInterpretation': 'Geobody Boundary Interpretation',
    'obj_GeologicUnitFeature': 'Geological Unit Feature',
    'obj_GeologicUnitInterpretation': 'Geological Unit Interpretation',
    'obj_RockFluidOrganizationInterpretation': 'Rock Fluid Organization Interpretation',
    'obj_RockFluidUnitFeature': 'Rock Fluid Unit Feature',
    'obj_RockFluidUnitInterpretation': 'Rock Fluid Unit Interpretation',
    # stratigraphy classes
    'obj_GlobalChronostratigraphicColumn': 'Global Chronostratigraphic Column',
    'obj_StratigraphicColumn': 'Stratigraphic Column',
    'obj_StratigraphicColumnRankInterpretation': 'Stratigraphic Column Rank Interpretation',
    'obj_StratigraphicOccurrenceInterpretation': 'Stratigraphic Occurrence Interpretation',
    'obj_StratigraphicUnitFeature': 'Stratigraphic Unit Feature',
    'obj_StratigraphicUnitInterpretation': 'Stratigraphic Unit Interpretation',
    # low level classes
    'obj_SubRepresentation': 'Sub Representation',
    'obj_RepresentationSetRepresentation': 'Representation Set',
    'obj_StringTableLookup': 'String Lookup Table',
    'obj_DoubleTableLookup': 'Double (real) Lookup Table',
    'obj_TriangulatedSetRepresentation': 'Triangulated Set',
    'obj_PolylineSetRepresentation': 'Poly Line Set',
    'obj_PolylineRepresentation': 'Poly Line',
    'obj_PointSetRepresentation': 'Point Set',
    'obj_PlaneSetRepresentation': 'Plane Set',
    'obj_PointsProperty': 'Points Property',
    'obj_RedefinedGeometryRepresentation': 'Redefined Geometry Representation',
    # doc props bits
    'application/vnd.openxmlformats-package.core-properties+xml': 'Documentation (core properties)',
    'application/x-extended-core-properties+xml': 'Documentation (extended properties)',
    # other
    'obj_Activity': 'Activity',
    'obj_ActivityTemplate': 'Activity Template',
    'obj_CommentProperty': 'Comment Property',
    'obj_CommentPropertySeries': 'Comment Property Series',
    'obj_RepresentationIdentitySet': 'Representation Identity Set',
    'obj_StreamlinesFeature': 'Streamlines Feature',
    'obj_StreamlinesRepresentation': 'Streamlines'
}


def readable_class(class_name):
    """Given a resqml object class name as a string, returns a more human readable string.

    argument:
       class_name (string): the resqml class name, eg. 'obj_IjkGridRepresentation'

    returns:
       a human readable version of the class name, eg. 'Grid (IJK)'
    """

    if class_name in class_dict.keys():
        return class_dict[class_name]
    if class_name.startswith('obj_'):
        return class_name[4:]
    elif 'obj_' + class_name in class_dict.keys():
        return class_dict['obj_' + class_name]
    return class_name



"""WellboreMarkerFrame class."""

# Nexus is a registered trademark of the Halliburton Company
# RMS and ROXAR are registered trademarks of Roxar Software Solutions AS, an Emerson company

import logging

log = logging.getLogger(__name__)

import numpy as np
import pandas as pd

import resqpy.olio.uuid as bu
import resqpy.olio.write_hdf5 as rwh5
import resqpy.olio.xml_et as rqet
import resqpy.organize as rqo
import resqpy.well
import resqpy.well._trajectory as rqt
import resqpy.well._wellbore_marker as rqwbm
import resqpy.well.well_utils as rqwu
from resqpy.olio.base import BaseResqpy
from resqpy.olio.xml_namespaces import curly_namespace as ns


class WellboreMarkerFrame(BaseResqpy):
    """Class to handle RESQML WellBoreMarkerFrameRepresentation objects.

    note:
       measured depth data must be in same crs as those for the related trajectory
    """

    resqml_type = 'WellboreMarkerFrameRepresentation'

    boundary_feature_dict = {
        'GeologicBoundaryKind': ['fault', 'geobody', 'horizon'],
        'FluidMarker': ['gas down to', 'gas up to', 'oil down to', 'oil up to', 'water down to', 'water up to'],
        'FluidContact': ['free water contact', 'gas oil contact', 'gas water contact', 'seal', 'water oil contact']
    }

    def __init__(self,
                 parent_model,
                 uuid = None,
                 trajectory_uuid = None,
                 title = None,
                 originator = None,
                 extra_metadata = None):
        """Creates a new wellbore marker frame object and optionally loads it from xml.

        arguments:
           parent_model (model.Model object): the model which the new wellbore marker frame belongs to
           uuid (uuid.UUID, optional): If given, loads from disk. Else, creates new.
           trajectory_uuid (uuid.UUID, optional): the uuid of the Trajectory object associated with the well;
              not used if uuid is not None
           title (str, optional): the citation title to use for a new wellbore marker frame;
              ignored if uuid is not None
           originator (str, optional): the name of the person creating the wellbore marker frame, defaults to login id;
              ignored if uuid is not None
           extra_metadata (dict, optional): string key, value pairs to add as extra metadata for the wellbore marker frame;
              ignored if uuid is not None

        returns:
           the newly created wellbore marker frame object

        :meta common:
        """
        self.trajectory_uuid = trajectory_uuid
        self.node_count = None
        self.node_mds = None
        self.marker_list = None
        self.trajectory = None

        super().__init__(model = parent_model,
                         uuid = uuid,
                         title = title,
                         originator = originator,
                         extra_metadata = extra_metadata)

        if self.trajectory_uuid is not None:
            self.trajectory = rqt.Trajectory(parent_model = self.model, uuid = self.trajectory_uuid)

    @classmethod
    def from_dataframe(cls,
                       parent_model,
                       trajectory_uuid,
                       dataframe,
                       md_col = 'MD',
                       boundary_feature_type_col = 'Boundary_Feature_Type',
                       marker_citation_title_col = 'Marker_Citation_Title',
                       interp_citation_title_col = 'Interp_Citation_Title',
                       create_organizing_objects_where_needed = True,
                       title = None,
                       originator = None,
                       extra_metadata = None):
        """Load wellbore marker frame data from a pandas data frame.

        arguments:
           parent_model (model.Model object): the model which the new blocked well belongs to
           dataframe: a pandas dataframe holding the wellbore marker frame data
           trajectory_uuid (uuid.UUID, optional): the uuid of the Trajectory object associated with the well;
           md_col (string, default 'MD'): the name of the column holding measured depth values
           boundary_feature_type_col (string, default 'Boundary_Feature_Type'): the name of the column holding the type of geologic,
            fluid or contact feature;
            e.g. "fault", "geobody", "horizon ", "gas/oil/water down to", "gas/oil/water up to",
            "free water contact", "gas oil contact", "gas water contact", "water oil contact", "seal"
           marker_citation_title_col (string, default 'Marker_Citation_Title'): the name of the column holding the marker citation title
           interp_citation_title_col (string, default 'Interp_Citation_Title'): the name of the column holding the interpretation
           create_organizing_objects_where_needed (bool, default True): if True, interpretation and feature objects will be created
              when needed if they do not already exist; if False, an exception will be raised if a required interpretation is missing
           citation title
           title (str, optional): the citation title to use for a new wellbore marker frame;
              ignored if uuid is not None
           originator (str, optional): the name of the person creating the wellbore marker frame, defaults to login id;
              ignored if uuid is not None
           extra_metadata (dict, optional): string key, value pairs to add as extra metadata for the wellbore marker frame;
              ignored if uuid is not None

        returns:
            the newly created wellbore marker frame object
        """

        # the interpretation_citation_column is not a mandatory column in the dataframe as FluidContact and FluidMarker
        # boundary features do not have corresponding interpretations.
        for col in [md_col, boundary_feature_type_col, marker_citation_title_col]:
            assert col in dataframe.columns
            # assert that none of the values is blank
            assert dataframe[col].isnull().sum() == 0, f'blank value found in {col}'

        # verify that the boundary feature types specified are valid
        for i, boundary_feature in enumerate(dataframe[boundary_feature_type_col]):
            assert boundary_feature.lower() in (["fault", "geobody", "horizon", "gas down to", "oil down to", "water down to",
                                                 "gas up to", "oil up to", "water up to", "free water contact", "gas oil contact",
                                                 "gas water contact", "water oil contact", "seal"]),\
                f"invalid boundary feature type specified in row {i}"

        # create a wellbore marker object for each of the rows of the dataframe
        marker_list = []
        for i, row in dataframe.iterrows():
            row_index = i
            row_marker_type = row[boundary_feature_type_col].lower()
            row_interp_uuid = WellboreMarkerFrame.__parse_dataframe_row_for_interp_uuid(
                parent_model = parent_model,
                row = row,
                row_marker_type = row_marker_type,
                interp_citation_title_col = interp_citation_title_col,
                create_organizing_objects_where_needed = create_organizing_objects_where_needed)
            row_marker_citation_title = row[marker_citation_title_col]
            row_wellbore_marker_object = rqwbm.WellboreMarker(parent_model = parent_model,
                                                              parent_frame = WellboreMarkerFrame,
                                                              marker_index = row_index,
                                                              marker_type = row_marker_type,
                                                              interpretation_uuid = row_interp_uuid,
                                                              title = row_marker_citation_title)
            marker_list.append(row_wellbore_marker_object)

        wellbore_marker_frame = cls(parent_model = parent_model,
                                    trajectory_uuid = trajectory_uuid,
                                    title = title,
                                    originator = originator,
                                    extra_metadata = extra_metadata)
        wellbore_marker_frame.node_count = len(dataframe)
        wellbore_marker_frame.node_mds = np.array(dataframe['MD'])
        wellbore_marker_frame.marker_list = marker_list
        # check that the number of measured depths matches the node count and the number of markers
        assert wellbore_marker_frame.node_count == wellbore_marker_frame.node_mds.shape[0] == len(marker_list)

        return wellbore_marker_frame

    @staticmethod
    def __parse_dataframe_row_for_interp_uuid(parent_model, row, row_marker_type, interp_citation_title_col,
                                              create_organizing_objects_where_needed):
        """Extract the uuid of the interpretation object that the marker relates to from a row in the dataframe."""

        row_marker_type = row_marker_type.lower()
        if row_marker_type in ['horizon', 'fault', 'geobody']:
            row_interp_type = 'obj_' + row_marker_type.capitalize() + 'Interpretation'
        else:
            row_interp_type = None
        try:
            if row_interp_type is not None:
                row_interp_citation_title = row[interp_citation_title_col]
                row_interp_uuid = parent_model.uuid(obj_type = row_interp_type, title = row_interp_citation_title)
                if row_interp_uuid is None:
                    if create_organizing_objects_where_needed:
                        row_interp_uuid = WellboreMarkerFrame._create_feature_and_interpretation(
                            parent_model, row_marker_type, row_interp_citation_title)
                    else:
                        raise ValueError(f'interpretation uuid cannot be found for title {row_interp_citation_title}')
            else:
                # no interpretation exists for the boundary feature
                row_interp_uuid = None
        except KeyError:  # no interpretation column in the dataframe
            row_interp_uuid = None

        return row_interp_uuid

    @staticmethod
    def _create_feature_and_interpretation(model, feature_type, feature_name):
        # includes create_xml() for created organisational objects
        feature_type = feature_type.lower()
        assert feature_type in ['fault', 'horizon', 'geobody']
        if feature_type == 'fault':
            tbf = rqo.TectonicBoundaryFeature(model, kind = 'fault', feature_name = feature_name)
        else:
            tbf = rqo.GeneticBoundaryFeature(model, kind = feature_type, feature_name = feature_name)
        tbf.create_xml(reuse = True)
        if feature_type == 'fault':
            fi = rqo.FaultInterpretation(model, tectonic_boundary_feature = tbf, is_normal = True)
            # todo: set is_normal correctly
        elif feature_type == 'horizon':
            fi = rqo.HorizonInterpretation(model, genetic_boundary_feature = tbf)
            # todo: support boundary relation list and sequence stratigraphy surface
        else:  # geobody boundary
            fi = rqo.GeobodyBoundaryInterpretation(model, genetic_boundary_feature = tbf)
        fi.create_xml(reuse = True)
        return fi.uuid

    def _load_from_xml(self):
        """Loads the wellbore marker frame object from an xml node (and associated hdf5 data).

        note:
           this method is not usually called directly
        """

        wellbore_marker_frame_root = self.root
        assert wellbore_marker_frame_root is not None

        self.trajectory_uuid = bu.uuid_from_string(
            rqet.find_nested_tags_text(wellbore_marker_frame_root, ['Trajectory', 'UUID']))
        self.trajectory = rqt.Trajectory(parent_model = self.model, uuid = self.trajectory_uuid)

        # list of Wellbore markers
        self.marker_list = []
        for i, tag in enumerate(rqet.list_of_tag(wellbore_marker_frame_root, 'WellboreMarker')):
            marker_obj = rqwbm.WellboreMarker(parent_model = self.model,
                                              parent_frame = self,
                                              marker_index = i,
                                              marker_node = tag)
            self.marker_list.append(marker_obj)

        self.node_count = rqet.find_tag_int(wellbore_marker_frame_root, 'NodeCount')
        rqwu.load_hdf5_array(self, rqet.find_tag(wellbore_marker_frame_root, 'NodeMd'), "node_mds", tag = 'Values')

        assert self.node_count == len(self.node_mds), 'node count does not match hdf5 array'
        assert len(self.marker_list) == self.node_count, 'wellbore marker list does not contain correct node count'

    def dataframe(self):
        """Returns a pandas dataframe with columns X, Y, Z, MD, Boundary_Feature_Type, Marker_Citation_Title,

        Interp_Citation_Title.
        """

        xyz = np.empty((self.node_count, 3))
        boundary_feature_type_list = []
        marker_citation_title_list = []
        interp_citation_title_list = []

        for i, marker_obj in enumerate(self.marker_list):
            boundary_feature_type = marker_obj.marker_type
            marker_citation_title = marker_obj.title
            interp_uuid = marker_obj.interpretation_uuid
            if interp_uuid is not None:
                interp_root = self.model.root_for_uuid(interp_uuid)
                interp_citation_title = self.model.title_for_root(root = interp_root)
            else:
                interp_citation_title = None
            xyz[i] = self.trajectory.xyz_for_md(self.node_mds[i])
            boundary_feature_type_list.append(boundary_feature_type)
            marker_citation_title_list.append(marker_citation_title)
            interp_citation_title_list.append(interp_citation_title)

        return pd.DataFrame({
            'X': xyz[:, 0],
            'Y': xyz[:, 1],
            'Z': xyz[:, 2],
            'MD': self.node_mds,
            'Boundary_Feature_Type': boundary_feature_type_list,
            'Marker_Citation_Title': marker_citation_title_list,
            'Interp_Citation_Title': interp_citation_title_list
        })

    def create_xml(self,
                   ext_uuid = None,
                   add_as_part = True,
                   add_relationships = True,
                   title = 'wellbore marker framework',
                   originator = None):
        """Creates the xml tree for this wellbore marker frame and optionally adds as a part to the model."""

        assert type(add_as_part) is bool

        if not self.title:
            self.title = title

        if ext_uuid is None:
            ext_uuid = self.model.h5_uuid()

        wbm_node = super().create_xml(originator = originator, add_as_part = False)
        node_count, nodeMd, md_values_node = self.__add_sub_elements_to_root_node(wbm_node = wbm_node)
        self.model.create_hdf5_dataset_ref(ext_uuid, self.uuid, 'Mds', root = md_values_node)

        if self.trajectory is not None:
            traj_root = self.trajectory.root
            self.model.create_ref_node('Trajectory',
                                       rqet.find_tag(rqet.find_tag(traj_root, 'Citation'), 'Title').text,
                                       bu.uuid_from_string(traj_root.attrib['uuid']),
                                       content_type = 'obj_WellboreTrajectoryRepresentation',
                                       root = wbm_node)
        else:
            log.error('trajectory object is missing and must be included')

        # fill wellbore marker
        if self.marker_list is not None:
            for marker in self.marker_list:
                title = self.__get_wbm_node_title(marker)
                wbm_node_obj = marker.create_xml(parent_node = wbm_node, title = title)
                assert wbm_node_obj is not None

        # add as part
        self.__add_as_part_and_add_relationships(wbm_node = wbm_node,
                                                 ext_uuid = ext_uuid,
                                                 add_as_part = add_as_part,
                                                 add_relationships = add_relationships)

        return wbm_node

    def __get_wbm_node_title(self, marker):
        """ Generate the title of the newly created wellbore marker object node."""

        well_name = self.trajectory.well_name
        for k, v in WellboreMarkerFrame.boundary_feature_dict.items():
            if marker.marker_type in v:
                boundary_feature = k
                break
        interp_title = self.model.title_for_root(root = self.model.root_for_uuid(uuid = marker.interpretation_uuid))
        title = ' '.join([x for x in [well_name, boundary_feature, marker.marker_type, interp_title] if x is not None])
        return title

    def __add_sub_elements_to_root_node(self, wbm_node):
        """Appends sub-elements to the WellboreMarkerFrame object's root node."""

        nodeCount = rqet.SubElement(wbm_node, ns['resqml2'] + 'NodeCount')
        nodeCount.set(ns['xsi'] + 'type', ns['xsd'] + 'positiveInteger')
        nodeCount.text = str(self.node_count)

        nodeMd = rqet.SubElement(wbm_node, ns['resqml2'] + 'NodeMd')
        nodeMd.set(ns['xsi'] + 'type', ns['resqml2'] + 'DoubleHdf5Array')
        nodeMd.text = rqet.null_xml_text

        md_values_node = rqet.SubElement(nodeMd, ns['resqml2'] + 'Values')
        md_values_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'Hdf5Dataset')
        md_values_node.text = rqet.null_xml_text

        return nodeCount, nodeMd, md_values_node

    def __add_as_part_and_add_relationships(self, wbm_node, ext_uuid, add_as_part, add_relationships):
        """Add the newly created WellboreMarkerFrame object's root node as a part in the model

        and add reciprocal relationships.
        """

        if add_as_part:
            self.model.add_part('obj_WellboreMarkerFrameRepresentation', self.uuid, wbm_node)

            if add_relationships:
                self.model.create_reciprocal_relationship(wbm_node, 'destinationObject', self.trajectory.root,
                                                          'sourceObject')
                ext_part = rqet.part_name_for_object('obj_EpcExternalPartReference', ext_uuid, prefixed = False)
                ext_node = self.model.root_for_part(ext_part)
                self.model.create_reciprocal_relationship(wbm_node, 'mlToExternalPartProxy', ext_node,
                                                          'externalPartProxyToMl')
                if self.marker_list is not None:  # TODO: confirm that you can have an empty wellbore marker frame with no marker list
                    for marker in self.marker_list:
                        interp_root = self.model.root_for_uuid(uuid = marker.interpretation_uuid)
                        self.model.create_reciprocal_relationship(wbm_node, 'destinationObject', interp_root,
                                                                  'sourceObject')

    def write_hdf5(self, file_name = None, mode = 'a'):
        """Writes the hdf5 array associated with this object (the measured depth data).

        arguments:
           file_name (string): the name of the hdf5 file, or None, in which case the model's default will be used
           mode (string, default 'a'): the write mode for the hdf5, either 'w' or 'a'
        """

        h5_reg = rwh5.H5Register(self.model)
        h5_reg.register_dataset(self.uuid, 'Mds', self.node_mds)
        h5_reg.write(file = file_name, mode = mode)

    def find_marker_index_from_interp(
            self, interpretation_uuid):  # TODO: return marker index from interp uuid, else return None
        """Find wellbore marker index by interpretation uuid.

        arguments:
           interpretation_uuid (uuid.UUID or string): the uuid of the interpretation object of interest

        returns:
           integer indicating the associated marker's position in self.marker_list

        note:
           if no marker is found for the interpretation uuid, None is returned
        """

        if type(interpretation_uuid) is str:
            interpretation_uuid = bu.uuid_from_string(interpretation_uuid)

        for i, marker in enumerate(self.marker_list):
            if bu.matching_uuids(marker.interpretation_uuid, interpretation_uuid):
                return i

        return None

    def find_marker_from_index(self, idx):
        """Returns wellbore marker by index.

        arguments:
           idx (int): position of the marker in the wellbore marker list

        returns:
           marker (WellboreMarker object)
        """
        try:
            found_marker = self.marker_list[idx]
        except IndexError:
            found_marker = None

        return found_marker



"""Class for a collection of well logs"""

import logging

log = logging.getLogger(__name__)

import numpy as np
import pandas as pd
import lasio
from datetime import datetime

import resqpy.property
import resqpy.weights_and_measures as bwam

import resqpy.property.property_collection as rqp_pc
import resqpy.property.property_common as rqp_c
import resqpy.property.well_log as rqp_wl


class WellLogCollection(rqp_pc.PropertyCollection):
    """Class for RESQML Property collection for a Wellbore Frame (ie well logs), inheriting from PropertyCollection."""

    def __init__(self, frame = None, property_set_root = None, realization = None):
        """Creates a new property collection related to a wellbore frame.

        arguments:
           frame (well.WellboreFrame object, optional): must be present unless creating a completely blank, empty collection.
              See :class:`resqpy.well.WellboreFrame`
           property_set_root (optional): if present, the collection is populated with the properties defined in the xml tree
              of the property set; frame must not be None when using this argument
           realization (integer, optional): if present, the single realisation (within an ensemble) that this collection is for;
              if None, then the collection is either covering a whole ensemble (individual properties can each be flagged with a
              realisation number), or is for properties that do not have multiple realizations

        returns:
           the new WellLogCollection object

        note:
           usually a wellbore frame should be passed, however a completely blank collection may be created prior to using
           collection inheritance methods to populate from another collection, in which case the frame can be lazily left
           as None here;
           for actual well logs, the realization argument will usually be None; for synthetic logs created from an ensemble
           it may be of use
        """

        if property_set_root is not None:
            super().__init__(support = frame, property_set_root = property_set_root, realization = realization)
        else:
            super().__init__()
            if frame is not None:
                self.set_wellbore_frame(frame)
                fpc = frame.extract_property_collection()
                self.inherit_parts_selectively_from_other_collection(fpc,
                                                                     realization = realization,
                                                                     indexable = 'nodes')

    def add_log(self, title, data, unit, discrete = False, realization = None, write = True, source_info = ''):
        """Add a well log to the collection, and optionally save to HDF / XML.

        note:
           If write=False, the data are not written to the model and are saved to be written later.
           To write the data, you can subsequently call::

              logs.write_hdf5_for_imported_list()
              logs.create_xml_for_imported_list_and_add_parts_to_model()

        arguments:
           title (str): Name of log, typically the mnemonic
           data (array-like): log data to write. Must have same length as frame MDs
           unit (str): Unit of measure
           discrete (bool): by default False, i.e. continuous
           realization (int): If given, assign data to a realisation.
           write (bool): If True, write XML and HDF5.
           source_info (str): curve description or other human readable text

        returns:
           uuids: list of uuids of newly added properties. Only returned if write=True.
        """
        # Validate
        if self.support is None:
            raise ValueError('Supporting WellboreFrame not present')
        if len(data) != self.support.node_count:
            raise ValueError(f'Data mismatch: data length={len(data)}, but MD node count={self.support.node_count}')

        # Infer valid RESQML properties
        # TODO: Store orginal unit somewhere if it's not a valid RESQML unit
        uom = bwam.rq_uom(unit)
        property_kind, facet_type, facet = rqp_c.infer_property_kind(title, uom)

        # Add to the "import list"
        self.add_cached_array_to_imported_list(
            cached_array = np.array(data),
            source_info = source_info,
            keyword = title,
            discrete = discrete,
            uom = uom,
            property_kind = property_kind,
            facet_type = facet_type,
            facet = facet,
            realization = realization,
        )

        if write:
            self.write_hdf5_for_imported_list()
            return self.create_xml_for_imported_list_and_add_parts_to_model()
        else:
            return None

    def iter_logs(self):
        """Generator that yields component Log objects.

        Yields:
           instances of :class:`resqpy.property.WellLog` .

        Example::

           for log in log_collection.logs():
              print(log.title)
        """

        return (rqp_wl.WellLog(collection = self, uuid = uuid) for uuid in self.uuids())

    def to_df(self, include_units = False):
        """Return pandas dataframe of log data.

        arguments:
           include_units (bool): include unit in column names
        """

        assert self.support is not None

        # Get MD values
        md_values = self.support.node_mds
        assert md_values.ndim == 1, 'measured depths not 1D numpy array'

        # Get logs
        data = {}
        for log in self.iter_logs():

            col_name = log.title
            if include_units and log.uom:
                col_name += f' ({log.uom})'

            values = log.values()
            if values.ndim > 1:
                raise NotImplementedError('Multidimensional logs not yet supported in pandas')

            data[col_name] = values

        df = pd.DataFrame(data = data, index = md_values)
        return df

    def to_las(self):
        """Return a lasio.LASFile object, which can then be written to disk.

        Example::

           las = collection.to_las()
           las.write('example_logs.las', version=2)
        """
        las = lasio.LASFile()

        las.well.WELL = str(self.support.wellbore_interpretation.title)
        las.well.DATE = datetime.today().strftime('%Y-%m-%d %H:%M:%S')

        # todo: Get UWI from somewhere
        # las.well.UWI = uwi

        # Lookup depths from associated WellboreFrame and Trajectory
        md_values = self.support.node_mds
        md_unit = self.support.trajectory.md_uom

        # Measured depths should be first column in LAS file
        # todo: include datum information in description
        las.append_curve('MD', md_values, unit = md_unit)

        for well_log in self.iter_logs():
            name = well_log.title
            unit = well_log.uom
            values = well_log.values()
            if values.ndim > 1:
                raise NotImplementedError('Multidimensional logs not yet supported in pandas')
            assert len(values) > 0
            log.debug(f"Writing log {name} of length {len(values)} and shape {values.shape}")
            las.append_curve(name, values, unit = unit, descr = None)
        return las

    def set_wellbore_frame(self, frame):
        """Sets the supporting representation object for the property collection to be the given wellbore frame object.

        note:
           this method does not need to be called if the wellbore frame was identified at the time the collection
           was initialised
        """

        self.set_support(support = frame)



"""relperm.py: class for dataframes of relative permeability data as RESQML objects.

Note that this module uses the obj_Grid2dRepresentation class in a way that was not envisaged
when the RESQML standard was defined; software that does not use resqpy is unlikely to be
able to do much with data stored in this way.
"""

# Nexus is a registered trademark of Halliburton

import logging

log = logging.getLogger(__name__)

import os
import numpy as np
import pandas as pd

from resqpy.olio.dataframe import DataFrame
import resqpy.olio.xml_et as rqet


class RelPerm(DataFrame):
    """Class for storing and retrieving a pandas dataframe of relative permeability data.

    note:
       inherits from DataFrame class
    """

    def __init__(self,
                 model,
                 uuid = None,
                 df = None,
                 uom_list = None,
                 realization = None,
                 phase_combo = None,
                 low_sal = False,
                 table_index = None,
                 title = 'relperm_table',
                 column_lookup_uuid = None,
                 uom_lookup_uuid = None):
        """Create a new RelPerm object from either a previously stored object or a pandas dataframe.

        arguments:
           phase_combo (str, optional): the combination of phases whose relative permeability behaviour is described.
              Options include 'water-oil', 'gas-oil' and 'gas-water'
           low_sal (boolean, optional): if True, indicates that the water-oil table contains the low-salinity data for
              relative permeability and capillary pressure
           table_index (int, optional): the index of the relative permeability
           table when multiple relative permeability tables are present. Note, indices should start at 1.

        note:
           see DataFrame class docstring for details of other arguments
        """

        # check that either a uuid OR dataframe has been provided
        if df is None and uuid is None:
            raise ValueError('either a uuid or a dataframe must be provided')

        # check extra_metadata for arguments if uuid is provided
        if uuid is not None:
            model_root = model.root(uuid = uuid)
            uuid_metadata_dict = rqet.load_metadata_from_xml(model_root)
            self.phase_combo = uuid_metadata_dict['phase_combo']
            self.low_sal = uuid_metadata_dict['low_sal']
            self.table_index = uuid_metadata_dict['table_index']

        else:
            # check that 'phase_combo' parameter is valid
            self.phase_combo = phase_combo
            self.low_sal = low_sal
            self.table_index = table_index

            df.columns = [x.capitalize() for x in df.columns]
            if 'Pc' in df.columns and df.columns[-1] != 'Pc':
                raise ValueError('Pc', 'capillary pressure data should be in the last column of the dataframe')

            processed_phase_combo_checks = {
                ('oil', 'water'): self.__water_oil_error_check,
                ('gas', 'oil'): self.__gas_oil_error_check,
                ('gas', 'water'): self.__gas_water_error_check,
                ('None',): self.__no_phase_combo_error_check
            }

            processed_phase_combo = tuple(sorted(set([x.strip() for x in str(self.phase_combo).split('-')])))
            if processed_phase_combo not in processed_phase_combo_checks.keys():
                raise ValueError('invalid phase_combo provided')
            # check that table_index is >= 1
            if table_index is not None and table_index < 1:
                raise ValueError('table_index cannot be less than 1')
            # check that the column names and order are as expected
            processed_phase_combo_checks.get(processed_phase_combo)(df)
            # ensure that missing capillary pressure values are stored as np.nan
            if 'Pc' in df.columns:
                df['Pc'].replace('None', np.nan, inplace = True)
            # convert all values in the dataframe to numeric type
            df[df.columns] = df[df.columns].apply(pd.to_numeric, errors = 'coerce')
            # ensure that no other column besides Pc has missing values
            cols_no_pc = [x for x in df.columns if 'Pc' != x]
            for col in cols_no_pc:
                if (df[col].isnull().sum() > 0) or ('None' in list(df[col])):
                    raise Exception(f'missing values found in {col} column')

            # check that Sw, Kr and Pc values are monotonic and that the Sw and Kr values are within the range 0-1
            sat_col = [x for x in df.columns if x[0] == 'S'][0]
            relp_corr_col = [x for x in df.columns if x == 'Kr' + sat_col[-1]][0]
            relp_opp_col = [x for x in df.columns if (x[0:2] == 'Kr') & (x[-1] != sat_col[-1])][0]
            if not (df[sat_col].is_monotonic and df[relp_corr_col].is_monotonic and df[
                relp_opp_col].is_monotonic_decreasing) and not \
                    (df[sat_col].is_monotonic_decreasing and df[relp_corr_col].is_monotonic_decreasing and
                     df[relp_opp_col].is_monotonic):
                raise ValueError(f'{sat_col, relp_corr_col, relp_opp_col} combo is not monotonic')
            if 'Pc' in df.columns:
                if not df['Pc'].dropna().is_monotonic and not df['Pc'].dropna().is_monotonic_decreasing:
                    raise ValueError('Pc values are not monotonic')
            for col in ['Sw', 'Sg', 'So', 'Krw', 'Krg', 'Kro']:
                if col in df.columns:
                    if df[col].min() < 0 or df[col].max() > 1:
                        raise ValueError(f'{col} is not within the range 0-1')

        super().__init__(model,
                         uuid = uuid,
                         df = df,
                         uom_list = uom_list,
                         realization = realization,
                         title = title,
                         column_lookup_uuid = column_lookup_uuid,
                         uom_lookup_uuid = uom_lookup_uuid,
                         extra_metadata = {
                             'relperm_table': 'true',
                             'phase_combo': self.phase_combo,
                             'low_sal': str(self.low_sal).lower(),
                             'table_index': str(self.table_index)
                         })

    @staticmethod
    def __error_check(expected_cols, sat_cols, df):
        if df.columns[0] not in sat_cols or len(set(df.columns).intersection(sat_cols)) != 1:
            raise ValueError('incorrect saturation column name and/or multiple saturation columns exist')
        if not set(df.columns).issubset(expected_cols):
            raise ValueError(f'incorrect column name(s) {set(df.columns).difference(expected_cols)}')

    def __water_oil_error_check(self, df):
        expected_cols = {'Sw', 'So', 'Krw', 'Kro', 'Pc'}
        sat_cols = {'Sw', 'So'}
        self.__error_check(expected_cols, sat_cols, df)

    def __gas_oil_error_check(self, df):
        expected_cols = {'Sg', 'So', 'Krg', 'Kro', 'Pc'}
        sat_cols = {'Sg', 'So'}
        self.__error_check(expected_cols, sat_cols, df)

    def __gas_water_error_check(self, df):
        expected_cols = {'Sg', 'Sw', 'Krg', 'Krw', 'Pc'}
        sat_cols = {'Sg', 'Sw'}
        self.__error_check(expected_cols, sat_cols, df)

    def __no_phase_combo_error_check(self, df):
        if df.columns[0] not in ['Sw', 'Sg', 'So'] or len(set(df.columns).intersection({'Sw', 'Sg', 'So'})) != 1:
            raise ValueError('incorrect saturation column name and/or multiple saturation columns exist')
        if set(df.columns).issubset({'Sw', 'So', 'Krw', 'Kro', 'Pc'}) and len(set(df.columns)) >= 3:
            self.phase_combo = 'water-oil'
        elif set(df.columns).issubset({'Sg', 'So', 'Krg', 'Kro', 'Pc'}) and len(set(df.columns)) >= 3:
            self.phase_combo = 'gas-oil'
        elif set(df.columns).issubset({'Sg', 'Sw', 'Krg', 'Krw', 'Pc'}) and len(set(df.columns)) >= 3:
            self.phase_combo = 'gas-water'
        else:
            raise Exception('unexpected number of columns and/or column headers')

    def interpolate_point(self, saturation, kr_or_pc_col):
        """Returns a tuple of the saturation value and the corresponding interpolated rel. perm. or cap. pressure value.

        arguments:
           saturation (float): the saturation at which the relative permeability or cap. pressure will be interpolated
           kr_or_pc_col (str): the column name of the parameter to be interpolated

        returns:
           tuple of float, the first element is the saturation and the second element is the interpolated value

        note:
           A simple linear interpolation is performed.
        """
        df = self.df.copy()
        if kr_or_pc_col.capitalize() not in df.columns or kr_or_pc_col.capitalize() == df.columns[0]:
            raise ValueError('incorrect column name provided for interpolation')
        if kr_or_pc_col == 'PC':
            df = df[df['PC'].notnull()]
        sat_col = df.columns[0]
        # ensure that the saturation values are monotonically increasing
        df = df.sort_values(by = sat_col)
        x = df[sat_col]
        y = df[kr_or_pc_col]
        x_new = saturation
        if x_new < x.min() or x_new > x.max():
            raise ValueError('saturation value is outside the interpolation range')
        y_new = np.interp(x_new, x, y)
        return saturation, y_new

    def df_to_text(self, filepath, filename):
        """Creates a text file from a dataframe of relative permeability and capillary pressure data.

        arguments:
           filepath (str): location where new text file is written to
           filename (str): name of the new text file

        returns:
           tuple of float, the first element is the saturation and the second element is the interpolated value

        note:
           Only Nexus compatible text files are currently supported. Text files that are compatible with other reservoir
              simulators may be supported in the future.
        """
        df = self.df.copy()
        ascii_file = os.path.join(filepath, filename + '.dat')
        df.columns = map(str.upper, df.columns)
        if {'KRW', 'KRO'}.issubset(set(df.columns)):
            if self.low_sal:
                table_name_keyword = 'WOTABLE (LOW_SAL)\n'
            else:
                table_name_keyword = 'WOTABLE\n'
            df.rename(columns = {'SW': 'SW', 'KRW': 'KRW', 'KRO': 'KROW', 'PC': 'PCWO'}, inplace = True)
        elif {'KRG', 'KRO'}.issubset(set(df.columns)):
            table_name_keyword = 'GOTABLE\n'
            df.rename(columns = {'SG': 'SG', 'KRG': 'KRG', 'KRO': 'KROG', 'PC': 'PCGO'}, inplace = True)
        elif {'KRW', 'KRW'}.issubset(set(df.columns)):
            table_name_keyword = 'GWTABLE\n'
            df.rename(columns = {'SG': 'SG', 'KRG': 'KRG', 'KRW': 'KRWG', 'PC': 'PCGW'}, inplace = True)
        else:
            raise Exception('incorrect rel. perm. column combination encountered')

        if filename + '.dat' not in os.listdir(filepath):
            with open(ascii_file, 'w') as f:
                f.write(table_name_keyword)
                df_str = df.to_string(na_rep = '', index = False)
                f.write(df_str)
                f.close()
                print(f'Created new DAT file: {filename} at {filepath}')
        else:
            with open(ascii_file, 'a') as f:
                f.write('\n\n')
                f.write(table_name_keyword)
                df_str = df.to_string(na_rep = '', index = False)
                f.write(df_str)
                f.close()
                print(f'Appended to DAT file: {filename} at {filepath}')

    def write_hdf5_and_create_xml(self):
        """Write relative permeability table data to hdf5 file and create xml for dataframe objects."""
        super().write_hdf5_and_create_xml()
        mesh_root = self.mesh.root
        # create an xml of extra metadata to indicate that this is a relative permeability table
        rqet.create_metadata_xml(mesh_root, self.extra_metadata)


def text_to_relperm_dict(relperm_data, is_file = True):
    """Return dict of dataframes with relative permeability and capillary pressure data and phase combinations.

    arguments:
       relperm_data (str): relative or full path of the text file to be processed or string of relative permeability data
       is_file (boolean): if True, indicates that a text file of relative permeability data has been provided. Default value is True

    returns:
       dict, each element in the dictionary contains a dataframe, with saturation and rel. permeability/capillary pressure
       data, and the phase combination being described

    note:
       Only Nexus compatible text files are currently supported. Text files from other reservoir simulators may be
          supported in the future.
    """
    if is_file:
        with open(relperm_data) as f:
            string_original = f.read()
    else:
        string_original = relperm_data
    # split the string based on newlines and remove any comments or other escape characters
    escapes = ''.join([chr(char) for char in range(1, 32)])
    string_formatted = [x.strip(escapes).split(' ') for x in filter(None, string_original.split('\n')) if '!' not in x]
    # remove all empty strings
    data = [list(filter(None, x)) for x in string_formatted]
    # get indices of start of each new relperm table based on Nexus keywords
    table_start_positions = [
        i for i, l in enumerate(data) if len({'WOTABLE', 'GOTABLE', 'GWTABLE'}.intersection(set(l))) == 1
    ]
    df_cols_dict = {
        'SW': 'Sw',
        'SG': 'Sg',
        'KRW': 'Krw',
        'KRG': 'Krg',
        'KROW': 'Kro',
        'KROG': 'Kro',
        'KRWG': 'Krw',
        'PCWO': 'Pc',
        'PCGO': 'Pc',
        'PCGW': 'Pc',
        'PC': 'Pc'
    }
    relperm_table_idx = 1
    rel_perm_dict = {}
    for i, l in enumerate(table_start_positions):
        key = 'relperm_table' + str(relperm_table_idx)
        rel_perm_dict[key] = {}
        relperm_table_idx += 1
        if 'WOTABLE' in data[l]:
            phase_combo = 'water-oil'
            rel_perm_dict[key]['phase_combo'] = phase_combo
        elif 'GOTABLE' in data[l]:
            phase_combo = 'gas-oil'
            rel_perm_dict[key]['phase_combo'] = phase_combo
        elif 'GWTABLE' in data[l]:
            phase_combo = 'gas-water'
            rel_perm_dict[key]['phase_combo'] = phase_combo
        else:
            raise Exception('incorrect table key word encountered')
        if i < (len(table_start_positions) - 1):
            table_end = table_start_positions[i + 1]
        else:
            table_end = len(data)
        table_cols = data[l + 1]
        table_rows = data[l + 2:table_end]
        df = pd.DataFrame(table_rows, columns = table_cols)
        df.columns = df.columns.map(df_cols_dict)
        sat_col = [x for x in df.columns if 'S' in x][0]
        df = df.apply(pd.to_numeric, errors = 'coerce')
        df = df[df[sat_col].notnull()]
        rel_perm_dict[key]['df'] = df
    return rel_perm_dict


def relperm_parts_in_model(model,
                           phase_combo = None,
                           low_sal = None,
                           table_index = None,
                           title = None,
                           related_uuid = None):
    """Returns list of part names within model that are representing RelPerm dataframe support objects.

    arguments:
       model (model.Model): the model to be inspected for dataframes
       phase_combo (str, optional): the combination of phases whose relative permeability behaviour is described.
          Options include 'water-oil', 'gas-oil', 'gas-water', 'oil-water', 'oil-gas' and 'water-gas'
       low_sal (boolean, optional): if True, indicates that the water-oil table contains the low-salinity data for
          relative permeability and capillary pressure
       table_index (int, optional): the index of the relative permeability table when multiple relative permeability
          tables are present. Note, indices should start at 1.
       title (str, optional): if present, only parts with a citation title exactly matching will be included
       related_uuid (uuid, optional): if present, only parts relating to this uuid are included

    returns:
       list of str, each element in the list is a part name, within model, which is representing the support for a RelPerm object
    """
    extra_metadata_orig = {
        'relperm_table': 'true',
        'phase_combo': phase_combo,
        'low_sal': low_sal,
        'table_index': table_index
    }
    extra_metadata = {k: str(v).lower() for k, v in extra_metadata_orig.items() if v is not None}
    df_parts_list = model.parts(obj_type = 'Grid2dRepresentation',
                                title = title,
                                extra = extra_metadata,
                                related_uuid = related_uuid)

    return df_parts_list



"""_polyline_set.py: Resqml polyline set module."""

import logging

log = logging.getLogger(__name__)

import os
import numpy as np

import resqpy.crs as rcrs
import resqpy.lines
import resqpy.olio.simple_lines as rsl
import resqpy.olio.uuid as bu
import resqpy.olio.vector_utilities as vu
import resqpy.olio.write_hdf5 as rwh5
import resqpy.olio.xml_et as rqet
import resqpy.lines._common as rql_c
from resqpy.olio.xml_namespaces import curly_namespace as ns


class PolylineSet(rql_c._BasePolyline):
    """Class for RESQML polyline set representation."""

    resqml_type = 'PolylineSetRepresentation'

    def __init__(self,
                 parent_model,
                 uuid = None,
                 polylines = None,
                 irap_file = None,
                 charisma_file = None,
                 crs_uuid = None,
                 title = None,
                 originator = None,
                 extra_metadata = None):
        """Initialises a new PolylineSet object.

        arguments:
            parent_model (model.Model object): the model which the new PolylineSetRepresentation belongs to
            uuid (uuid.UUID, optional): the uuid of an existing RESQML PolylineSetRepresentation object from
                which to initialise this resqpy PolylineSet
            polylines (optional): list of polyline objects from which to build the polylineset
            irap_file (str, optional): the name of a file in irap format from which to import the polyline set
            charisma_file (str, optional): the name of a file in charisma format from which to import the polyline set
            crs_uuid (UUID, optional): required if loading from an IRAP or Charisma file
            title (str, optional): the citation title to use for a new polyline set;
                ignored if uuid is not None
            originator (str, optional): the name of the person creating the polyline set, defaults to login id;
                ignored if uuid is not None
            extra_metadata (dict, optional): string key, value pairs to add as extra metadata for the polyline set;
                ignored if uuid is not None

        returns:
            the newly instantiated PolylineSet object

        note:
            this initialiser does not perform any conversion between CRSes; if loading from a list of Polylines,
            they must all share a common CRS (which must also match crs_uuid if supplied)

        :meta common:
        """

        self.model = parent_model
        self.coordinates = None
        self.count_perpol = None
        self.polys = []
        self.rep_int_root = None
        self.save_polys = False
        self.boolnotconstant = None
        self.boolvalue = None
        self.closed_array = None
        self.crs_uuid = crs_uuid
        self.indices = None

        super().__init__(model = parent_model,
                         uuid = uuid,
                         title = title,
                         originator = originator,
                         extra_metadata = extra_metadata)

        if self.root is not None:
            return

        if polylines is not None:  # Create from list of polylines
            crs_list = [] if crs_uuid is None else [bu.uuid_from_string(crs_uuid)]
            for poly in polylines:
                crs_list.append(poly.crs_uuid)
            crs_set = set(crs_list)
            # following assumes consistent use of UUID or str
            assert len(crs_set) == 1, 'More than one CRS found in input polylines for polyline set'
            for crs_uuid in crs_set:
                self.crs_uuid = crs_uuid
                if self.crs_uuid is not None:
                    break
            self.polys = polylines
            # Setting the title of the first polyline given as the PolylineSet title
            if not self.title:
                if len(polylines) > 1:
                    self.title = f"{polylines[0].title} + {len(polylines)-1} polylines"
                else:
                    self.title = polylines[0].title

        elif irap_file is not None:  # Create from an input IRAP file
            if crs_uuid is None:
                log.warning(f'applying generic model CRS when loading polylines from IRAP file: {irap_file}')
            self._set_from_irap(irap_file)

        elif charisma_file is not None:
            if crs_uuid is None:
                log.warning(f'applying generic model CRS when loading polylines from Charisma file: {charisma_file}')
            self._set_from_charisma(charisma_file)

    def _load_from_xml(self):

        assert self.root is not None  # polyline set xml node specified
        root = self.root

        self.rep_int_root = self.model.referenced_node(rqet.find_tag(root, 'RepresentedInterpretation'))
        self.closed_array = np.empty((0,), dtype = bool)

        for patch_node in rqet.list_of_tag(root, 'LinePatch'):  # Loop over all LinePatches - likely just the one
            assert patch_node is not None  # Required field

            geometry_node = rqet.find_tag(patch_node, 'Geometry')
            assert geometry_node is not None  # Required field

            crs_root = self.model.referenced_node(rqet.find_tag(geometry_node, 'LocalCrs'))
            assert crs_root is not None  # Required field
            uuid = rqet.uuid_for_part_root(crs_root)
            assert uuid is not None  # Required field
            if self.crs_uuid is not None:
                assert bu.matching_uuids(self.crs_uuid, uuid), 'mixed CRS uuids in use in PolylineSet'
            self.crs_uuid = uuid

            closed_node = rqet.find_tag(patch_node, 'ClosedPolylines')
            assert closed_node is not None  # Required field
            # The ClosedPolylines could be a BooleanConstantArray, or a BooleanArrayFromIndexArray
            closed_array = self.get_bool_array(closed_node)

            count_node = rqet.find_tag(patch_node, 'NodeCountPerPolyline')
            rql_c.load_hdf5_array(self, count_node, 'count_perpol', tag = 'Values', dtype = 'int')

            points_node = rqet.find_tag(geometry_node, 'Points')
            assert points_node is not None  # Required field
            rql_c.load_hdf5_array(self, points_node, 'coordinates', tag = 'Coordinates')

            # Check that the number of bools aligns with the number of count_perpoly
            # Check that the total of the count_perpoly aligns with the number of coordinates
            assert len(self.count_perpol) == len(closed_array)
            assert np.sum(self.count_perpol) == len(self.coordinates)

            subpolys = self.convert_to_polylines(closed_array, self.count_perpol, self.coordinates, self.crs_uuid,
                                                 self.rep_int_root)
            # Check we have the right number of polygons
            assert len(subpolys) == len(self.count_perpol)

            # Remove duplicate coordinates and count arrays (exist in polylines now)
            # delattr(self,'coordinates')
            # delattr(self,'count_perpol')

            self.polys.extend(subpolys)
            self.closed_array = np.concatenate((self.closed_array, closed_array))

        self.bool_array_format()

    def _set_from_irap(self, irap_file):
        inpoints = rsl.read_lines(irap_file)
        self.count_perpol = []
        closed_array = []
        if not self.title:
            self.title = os.path.basename(irap_file).split(".")[0]
        for i, poly in enumerate(inpoints):
            if len(poly) > 1:  # Polylines must have at least 2 points
                self.count_perpol.append(len(poly))
                if vu.isclose(poly[0], poly[-1]):
                    closed_array.append(True)
                else:
                    closed_array.append(False)
                if i == 0:
                    self.coordinates = poly
                else:
                    self.coordinates = np.concatenate((self.coordinates, poly))
        self.count_perpol = np.array(self.count_perpol)
        if self.crs_uuid is None:  # If no crs_uuid is provided, assume the main model crs is valid
            self.crs_uuid = self.model.crs_uuid
        self.polys = self.convert_to_polylines(closed_array, self.count_perpol, self.coordinates, self.crs_uuid,
                                               self.rep_int_root)

    def _set_from_charisma(self, charisma_file):
        with open(charisma_file) as f:
            inpoints = f.readlines()
        self.count_perpol = []
        closed_array = []
        if not self.title:
            self.title = os.path.basename(charisma_file).split(".")[0]
        for i, line in enumerate(inpoints):
            line = line.split()
            coord_entry = np.array([[float(line[3]), float(line[4]), float(line[5])]])
            if i == 0:
                self.coordinates = coord_entry
                stick = line[7]
                count = 1
            else:
                self.coordinates = np.concatenate((self.coordinates, coord_entry))
                count += 1
                if stick != line[7] or i == len(inpoints) - 1:
                    if count <= 2 and stick != line[7]:  # Line has fewer than 2 points
                        log.warning(f"Polylines must contain at least 2 points - ignoring point {self.coordinates[-2]}")
                        self.coordinates = np.delete(self.coordinates, -2, 0)  # Remove the second to last entry
                    else:
                        self.count_perpol.append(count - 1)
                        closed_array.append(False)
                    count = 1
                    stick = line[7]
        self.count_perpol = np.array(self.count_perpol)
        if self.crs_uuid is None:  # If no crs_uuid is provided, assume the main model crs is valid
            self.crs_uuid = self.model.crs_uuid
        self.polys = self.convert_to_polylines(closed_array, self.count_perpol, self.coordinates, self.crs_uuid,
                                               self.rep_int_root)

    def poly_index_containing_point_in_xy(self, p, mode = 'crossing'):
        """Returns the index of the first (closed) polyline containing point p in the xy plane, or None.

        :meta common:
        """

        assert mode in ['crossing', 'winding'], 'unrecognised mode when looking for polygon containing point'

        for i, poly in enumerate(self.polys):
            if not poly.isclosed:
                continue
            if poly.point_is_inside_xy(p, mode = mode):
                return i
        return None

    def create_xml(self,
                   ext_uuid = None,
                   add_as_part = True,
                   add_relationships = True,
                   title = None,
                   originator = None,
                   save_polylines = False):
        """Create xml from polylineset.

        args:
            save_polylines: If true, polylines are also saved individually

        :meta common:
        """

        if ext_uuid is None:
            ext_uuid = self.model.h5_uuid()

        self.save_polys = save_polylines

        if title:
            self.title = title
        if not self.title:
            self.title = 'polyline set'

        if self.save_polys:
            for poly in self.polys:
                poly.create_xml(ext_uuid, add_relationships = add_relationships, originator = originator)

        polyset = super().create_xml(add_as_part = False, originator = originator)

        if self.rep_int_root is not None:
            rep_int = self.rep_int_root
            if "FaultInterpretation" in str(rqet.content_type(rep_int)):
                content_type = 'obj_FaultInterpretation'
            else:
                content_type = 'obj_HorizonInterpretation'
            self.model.create_ref_node('RepresentedInterpretation',
                                       rqet.citation_title_for_node(rep_int),
                                       rep_int.attrib['uuid'],
                                       content_type = content_type,
                                       root = polyset)

        # We convert all Polylines to the CRS of the first Polyline in the set, so set this as crs_uuid

        patch = rqet.SubElement(polyset, ns['resqml2'] + 'LinePatch')
        patch.set(ns['xsi'] + 'type', ns['resqml2'] + 'PolylineSetPatch')
        patch.text = '\n'

        pindex = rqet.SubElement(patch, ns['resqml2'] + 'PatchIndex')
        pindex.set(ns['xsi'] + 'type', ns['xsd'] + 'nonNegativeInteger')
        pindex.text = '0'

        if self.boolnotconstant:

            # We have mixed data - use a BooleanArrayFromIndexArray
            closed = rqet.SubElement(patch, ns['resqml2'] + 'ClosedPolylines')
            closed.set(ns['xsi'] + 'type', ns['xsd'] + 'BooleanArrayFromIndexArray')
            closed.text = '\n'

            bool_val = rqet.SubElement(closed, ns['resqml2'] + 'IndexIsTrue')
            bool_val.set(ns['xsi'] + 'type', ns['xsd'] + 'boolean')
            bool_val.text = str(not self.boolvalue).lower()

            count = rqet.SubElement(closed, ns['resqml2'] + 'Count')
            count.set(ns['xsi'] + 'type', ns['xsd'] + 'positiveInteger')
            count.text = str(len(self.count_perpol))

            ind_val = rqet.SubElement(closed, ns['resqml2'] + 'Indices')
            ind_val.set(ns['xsi'] + 'type', ns['resqml2'] + 'IntegerHdf5Array')
            ind_val.text = '\n'

            iv_null = rqet.SubElement(ind_val, ns['resqml2'] + 'NullValue')
            iv_null.set(ns['xsi'] + 'type', ns['xsd'] + 'integer')
            iv_null.text = '-1'

            iv_values = rqet.SubElement(ind_val, ns['resqml2'] + 'Values')
            iv_values.set(ns['xsi'] + 'type', ns['eml'] + 'Hdf5Dataset')
            iv_values.text = '\n'

            self.model.create_hdf5_dataset_ref(ext_uuid, self.uuid, 'indices_patch0', root = iv_values)

        else:

            # All bools are the same - use a BooleanConstantArray
            closed = rqet.SubElement(patch, ns['resqml2'] + 'ClosedPolylines')
            closed.set(ns['xsi'] + 'type', ns['resqml2'] + 'BooleanConstantArray')
            closed.text = '\n'

            bool_val = rqet.SubElement(closed, ns['resqml2'] + 'Value')
            bool_val.set(ns['xsi'] + 'type', ns['xsd'] + 'boolean')
            bool_val.text = str(self.boolvalue).lower()

            count = rqet.SubElement(closed, ns['resqml2'] + 'Count')
            count.set(ns['xsi'] + 'type', ns['xsd'] + 'positiveInteger')
            count.text = str(len(self.count_perpol))

        count_pp = rqet.SubElement(patch, ns['resqml2'] + 'NodeCountPerPolyline')
        count_pp.set(ns['xsi'] + 'type', ns['resqml2'] + 'IntegerHdf5Array')
        count_pp.text = '\n'

        null = rqet.SubElement(count_pp, ns['resqml2'] + 'NullValue')
        null.set(ns['xsi'] + 'type', ns['xsd'] + 'integer')
        null.text = '0'

        count_val = rqet.SubElement(count_pp, ns['resqml2'] + 'Values')
        count_val.set(ns['xsi'] + 'type', ns['eml'] + 'Hdf5Dataset')
        count_val.text = '\n'

        self.model.create_hdf5_dataset_ref(ext_uuid, self.uuid, 'NodeCountPerPolyline_patch0', root = count_val)

        geom = rqet.SubElement(patch, ns['resqml2'] + 'Geometry')
        geom.set(ns['xsi'] + 'type', ns['resqml2'] + 'PointGeometry')
        geom.text = '\n'

        self.model.create_crs_reference(crs_uuid = self.crs_uuid, root = geom)

        points = rqet.SubElement(geom, ns['resqml2'] + 'Points')
        points.set(ns['xsi'] + 'type', ns['resqml2'] + 'Point3dHdf5Array')
        points.text = '\n'

        coords = rqet.SubElement(points, ns['resqml2'] + 'Coordinates')
        coords.set(ns['xsi'] + 'type', ns['eml'] + 'Hdf5Dataset')
        coords.text = '\n'

        self.model.create_hdf5_dataset_ref(ext_uuid, self.uuid, 'points_patch0', root = coords)

        if add_as_part:
            self.model.add_part('obj_PolylineSetRepresentation', self.uuid, polyset)
            if add_relationships:
                crs_root = self.model.root_for_uuid(self.crs_uuid)
                self.model.create_reciprocal_relationship(polyset, 'destinationObject', crs_root, 'sourceObject')
                if self.rep_int_root is not None:  # Optional
                    self.model.create_reciprocal_relationship(polyset, 'destinationObject', self.rep_int_root,
                                                              'sourceObject')
                if self.save_polys:
                    for poly in self.polys:
                        self.model.create_reciprocal_relationship(polyset, 'destinationObject', poly.root_node,
                                                                  'sourceObject')

                ext_part = rqet.part_name_for_object('obj_EpcExternalPartReference', ext_uuid, prefixed = False)
                ext_node = self.model.root_for_part(ext_part)
                self.model.create_reciprocal_relationship(polyset, 'mlToExternalPartProxy', ext_node,
                                                          'externalPartProxyToMl')

        return polyset

    def write_hdf5(self, file_name = None, mode = 'a', save_polylines = False):
        """Create or append the coordinates, counts and indices hdf5 arrays to hdf5 file.

        :meta common:
        """

        if self.uuid is None:
            self.uuid = bu.new_uuid()
        self.combine_polylines(self.polys)
        self.bool_array_format()
        self.save_polys = save_polylines
        if self.save_polys:
            for poly in self.polys:
                poly.write_hdf5(file_name)

        h5_reg = rwh5.H5Register(self.model)
        h5_reg.register_dataset(self.uuid, 'points_patch0', self.coordinates)
        h5_reg.register_dataset(self.uuid, 'NodeCountPerPolyline_patch0', self.count_perpol.astype(np.int32))
        if self.boolnotconstant:
            h5_reg.register_dataset(self.uuid, 'indices_patch0', np.array(self.indices, dtype = np.int32))
        h5_reg.write(file_name, mode = mode)

    def get_bool_array(self, closed_node):
        # TODO: check for other permissible forms of the abstract boolean array
        """Returns a boolean array using details in the xml node location.

        arguments:
            closed_node (xml node): the node under which the boolean array information sits

        returns:
            1D numpy bool array set True for those pplylines within the set which are marked as closed
        """
        # if type of boolean array is BooleanConstantArray, uses the array value and count to generate
        # the array; if type of boolean array is BooleanArrayFromIndexArray, finds the "other" value
        # bool and indices of the "other" values, and inserts these into an array opposite to the main bool
        flavour = rqet.node_type(closed_node)
        if flavour == 'BooleanConstantArray':
            count = rqet.find_tag_int(closed_node, 'Count')
            value = rqet.bool_from_text(rqet.node_text(rqet.find_tag(closed_node, 'Value')))
            return np.full((count,), value, dtype = bool)
        elif flavour == 'BooleanArrayFromIndexArray':
            count = rqet.find_tag_int(closed_node, 'Count')
            indices_node = rqet.find_tag(closed_node, 'Indices')
            assert indices_node is not None
            indices_arr = rql_c.load_hdf5_array(self, indices_node, 'indices_arr', tag = 'Values', dtype = 'int')
            istrue = rqet.bool_from_text(rqet.node_text(rqet.find_tag(closed_node, 'IndexIsTrue')))
            out = np.full((count,), not istrue, dtype = bool)
            out[indices_arr] = istrue
            return out
        raise ValueError(f'unrecognised closed array xml node type: {flavour}')

    def convert_to_polylines(self,
                             closed_array = None,
                             count_perpol = None,
                             coordinates = None,
                             crs_uuid = None,
                             rep_int_root = None):
        """Returns a list of Polylines objects from a PolylineSet.

        note:
            all arguments are optional and by default the data will be taken from self

        args:
            closed_array: array containing a bool for each polygon in if it is open (False) or closed (True)
            count_perpol: array containing a list of polygon "lengths" for each polygon
            coordinates: array containing coordinates for all the polygons
            crs_uuid: crs_uuid for polylineset
            rep_int_root: represented interpretation root (optional)

        returns:
            list of polyline objects

        :meta common:
        """

        from resqpy.lines._polyline import Polyline

        if count_perpol is None:
            count_perpol = self.count_perpol
        if closed_array is None:
            closed_array = np.zeros(len(count_perpol), dtype = bool)
            closed_node = rqet.find_nested_tags(self.root, ['LinePatch', 'ClosedPolylines'])
            if closed_node is not None:
                closed_array[:] = self.get_bool_array(closed_node)
        if coordinates is None:
            coordinates = self.coordinates
        if crs_uuid is None:
            crs_uuid = self.crs_uuid
        if rep_int_root is None:
            rep_int_root = self.rep_int_root
        polys = []
        count = 0
        for i in range(len(count_perpol)):
            if i != len(count_perpol) - 1:
                subset = coordinates[count:int(count_perpol[i]) + count].copy()
            else:
                subset = coordinates[count:int(count_perpol[i]) + count + 1].copy()
            if vu.isclose(subset[0], subset[-1]):
                isclosed = True
            else:
                isclosed = closed_array[i]
            count += int(count_perpol[i])
            subtitle = f"{self.title} {i+1}"
            polys.append(
                Polyline(self.model,
                         is_closed = isclosed,
                         set_coord = subset,
                         set_crs = crs_uuid,
                         title = subtitle,
                         rep_int_root = rep_int_root))

        return polys

    def combine_polylines(self, polylines):
        """Combines the isclosed boolean array, coordinates and count data for a list of polyline objects.

        args:
            polylines: list of polyline objects
        """

        self.count_perpol = []
        self.closed_array = []

        for poly in polylines:
            if poly == polylines[0]:
                master_crs = rcrs.Crs(self.model, uuid = poly.crs_uuid)
                self.crs_uuid = poly.crs_uuid
                self.coordinates = poly.coordinates.copy()
            else:
                curr_crs = rcrs.Crs(self.model, uuid = poly.crs_uuid)
                if not curr_crs.is_equivalent(master_crs):
                    shifted = curr_crs.convert_array_to(master_crs, poly.coordinates)
                    self.coordinates = np.concatenate((self.coordinates, shifted))
                else:
                    self.coordinates = np.concatenate((self.coordinates, poly.coordinates))

            self.closed_array.append(poly.isclosed)
            self.count_perpol.append(int(len(poly.coordinates)))

        self.count_perpol = np.array(self.count_perpol)

        assert len(self.closed_array) == len(self.count_perpol)
        assert np.sum(self.count_perpol) == len(self.coordinates)

        self.polys = polylines

    def bool_array_format(self):
        """Determines an appropriate output boolean array format depending on the closed_array bools.

        self.boolnotconstant - set to True if all are not open or all closed
        self.boolvalue - value of isclosed for all polylines, or for the majority of polylines if mixed
        self.indices - array of indices where the values are not self.boolvalue, if the polylines are mixed
        """

        assert self.closed_array is not None
        self.indices = []
        self.boolnotconstant = False
        if all(self.closed_array):
            self.boolvalue = True
        elif not any(self.closed_array):
            self.boolvalue = False
        else:
            if np.count_nonzero(self.closed_array) > (len(self.closed_array) // 2):
                self.boolvalue = True
                for i, val in enumerate(self.closed_array):
                    if not val:
                        self.indices.append(i)
            else:
                self.boolvalue = False
                for i, val in enumerate(self.closed_array):
                    if val:
                        self.indices.append(i)
            self.boolnotconstant = True

    def set_interpretation_root(self, rep_int_root, recursive = True):
        """Updates the rep_int_root for the polylineset.

        args:
            rep_int_root: new rep_int_root
            recursive: boolean, if true will update individual polys with same root
        """

        self.rep_int_root = rep_int_root

        if recursive:
            for poly in self.polys:
                poly.rep_int_root = rep_int_root

    def convert_to_irap(self, file_name):
        """Output an irap file from a polyline set.

        If file_name exists, it will be overwritten.

        args:
            file_name: output file name for polyline set representation
        """

        end_of_line = np.array([[999.0, 999.0, 999.0]])
        for poly in self.polys:
            if poly == self.polys[0]:
                out_coords = poly.coordinates
            else:
                out_coords = np.concatenate((out_coords, poly.coordinates))
            out_coords = np.concatenate((out_coords, end_of_line))
        np.savetxt(file_name, out_coords, delimiter = ' ')

    def convert_to_charisma(self, file_name):
        """Output to Charisma fault sticks from a polyline set.

        If file_name exists, it will be overwritten.

        args:
            file_name: output file name for polyline set representation
        """

        faultname = self.title.replace(" ", "_") if self.title else 'fault'
        lines = []
        for i, poly in enumerate(self.polys):
            for point in poly.coordinates:
                lines.append(f"INLINE-\t0\t0\t{point[0]}\t{point[1]}\t{point[2]}\t{faultname}\t{i+1}\n")
        with open(file_name, 'w') as f:
            for item in lines:
                f.write(item)



"""Cuda based grid surface intersection functionality for GPU processing.

note:
   use of this module requires accessible GPUs and the corresponding numba.cuda and cupy packages to be installed
"""

import logging

log = logging.getLogger(__name__)

import math as maths
import numpy as np
from typing import Tuple, Optional, Dict
import threading

import numba  # type: ignore
from numba import cuda  # type: ignore
from numba.cuda.cudadrv.devicearray import DeviceNDArray  # type: ignore
import cupy  # type: ignore

import resqpy.crs as rqc
import resqpy.grid as grr
import resqpy.fault as rqf
import resqpy.property as rqp
import resqpy.grid_surface._find_faces as rgs_ff
import resqpy.olio.uuid as bu

compiler_lock = threading.Lock()  # Numba compiler is not threadsafe


# cuda device wrappers for numpy functions
@cuda.jit(device = True)
def _cross_d(A: DeviceNDArray, B: DeviceNDArray, c: DeviceNDArray):
    c[0] = A[1] * B[2] - A[2] * B[1]
    c[1] = A[2] * B[0] - A[0] * B[2]
    c[2] = A[0] * B[1] - A[1] * B[0]


@cuda.jit(device = True)
def _negative_d(v: DeviceNDArray, nv: DeviceNDArray):
    for d in range(v.shape[0]):
        nv[d] = numba.float32(-1.0) * v[d]


@cuda.jit(device = True)
def _dot_d(v1: DeviceNDArray, v2: DeviceNDArray, prod: DeviceNDArray):
    prod[0] = 0.0
    for d in range(v1.shape[0]):
        prod[0] += v1[d] * v2[d]


@cuda.jit(device = True)
def _norm_d(v: DeviceNDArray, n: DeviceNDArray):
    n[0] = 0.0
    for dim in range(3):
        n[0] += v[dim]**2.0
    n[0] = maths.sqrt(n[0])


@cuda.jit
def project_polygons_to_surfaces(
    faces: DeviceNDArray,
    triangles: DeviceNDArray,
    axis: int,
    index1: int,
    index2: int,
    colx: int,
    coly: int,
    nx: int,
    ny: int,
    nz: int,
    dx: float,
    dy: float,
    dz: float,
    l_tol: float,
    t_tol: float,
    return_normal_vectors: bool,
    normals: DeviceNDArray,
    return_depths: bool,
    depths: DeviceNDArray,
    return_offsets: bool,
    offsets: DeviceNDArray,
    return_triangles: bool,
    triangle_per_face: DeviceNDArray,
):
    """Maps the projection of a 3D polygon to 2D grid surfaces along a given axis, using GPUs.

    arguments:
        faces (DeviceNDArray.bool): boolean array of each cell face that can represent the surface.
            nb. ordered k,j,i and sized (k,j,i)[axis] -= 1
        triangles (DeviceNDArray.float): ntriangles x naxis array containing (x,y,z) coordinates of each traingle.
        n_axis (int): number of cells in the axis.
        axis (int): axis number. Axis i is 0, j is 1, and k is 2.
        index1 (int): the first index. Axis i is 0, j is 0, and k is 1.
        index2 (int): the second index. Axis i is 1, j is 2, and k is 2.
        nx (int): number of points in x axis.
        ny (int): number of points in y axis.
        dx (float): cell's thickness along x-axis.
        dy (float): cell's thickness along y-axis.
        dz (float): cell's thickness along z-axis.
        l_tol (float, default 0.0): a fraction of the line length to allow for an intersection to be found
            just outside the segment.
        t_tol (float, default 0.0): a fraction of the triangle size to allow for an intersection to be found
            just outside the triangle.

    returns:
        void: modified faces array (INTENT OUT).
    """

    # define thread-local arrays used generally
    grid_nxyz = numba.cuda.local.array(3, numba.int32)
    grid_dxyz = numba.cuda.local.array(3, numba.float64)
    grid_nxyz[0] = nx
    grid_nxyz[1] = ny
    grid_nxyz[2] = nz
    grid_dxyz[0] = numba.float64(dx)
    grid_dxyz[1] = numba.float64(dy)
    grid_dxyz[2] = numba.float64(dz)
    # define thread-local arrays for section 3
    tp = numba.cuda.local.array((3, 3), numba.float64)
    line_p = numba.cuda.local.array(3, numba.float64)
    line_v = numba.cuda.local.array(3, numba.float64)
    p01 = numba.cuda.local.array(3, numba.float64)
    p02 = numba.cuda.local.array(3, numba.float64)
    lp_t0 = numba.cuda.local.array(3, numba.float64)
    norm = numba.cuda.local.array(3, numba.float64)
    line_rv = numba.cuda.local.array(3, numba.float64)
    tmp = numba.cuda.local.array(3, numba.float64)
    face_idx = numba.cuda.local.array(3, numba.int32)
    norm_idx = numba.cuda.local.array(3, numba.int32)
    xyz = numba.cuda.local.array(3, numba.float64)
    # scalars that must be returned from device functions must be mutable
    # => make them arrays
    u = numba.cuda.local.array(1, numba.float64)
    v = numba.cuda.local.array(1, numba.float64)
    denom = numba.cuda.local.array(1, numba.float64)
    t = numba.cuda.local.array(1, numba.float64)

    # extract useful dimension info
    n_axis = grid_nxyz[axis]  # get length of projection axis
    n_faces = faces.shape[2 - axis]  # n_faces == n_axis -1
    ntriangles = triangles.shape[0]

    #  cuda.grid(1) gives the thread index (blockIdx.x*blockDim.x + threadIdx.x)
    if cuda.grid(1) >= ntriangles:  # cuda.grid(1) evaluates to 1 int
        return  # this is actually unnecessary as the for-loop takes care of bounds

    # we have a set number of threads in a grid, so process each thread's
    # data and move it along to its next point in the array
    # Array:                          * * * * * * * * * * * * * * * * * * * * * * * * * * * * *|
    # > iteration 1- thread position: ^ ^ ^ ^ ^ ^ ^ ^                                          |        # cuda.grid(1)
    # > iteration 2- thread position:                 ^ ^ ^ ^ ^ ^ ^ ^                          |        # cuda.grid(1) + 1*cuda.gridsize(1)
    # > iteration 3- thread position:                                 ^ ^ ^ ^ ^ ^ ^ ^          |        # cuda.grid(1) + 2*cuda.gridsize(1)
    # > iteration 4- thread position:                                                 ^ ^ ^ ^ ^|x x x   # cuda.grid(1) + 3*cuda.gridsize(1)
    for triangle_num in range(cuda.grid(1), ntriangles, cuda.gridsize(1)):
        # the number of threads spawned should be enough to cover all triangles in one iteration
        # ...just imagine that this is the parallel section (like #pragma omp parallel)
        # 1. find triangle bounding box in this projection
        # 1a. get triangle under consideration
        # 1b. convert triangle-points coordinate to index

        for ver in range(3):  # for v in vertices
            for dim in range(3):  # for d in dimnensions
                tp[ver,
                   dim] = (numba.float64(triangles[triangle_num, ver, dim]) /
                           numba.float64(grid_dxyz[dim])) - numba.float64(0.5)  # get index of each aligned triangle

        # 1c. find triangle bounding box
        min_tpx = max(maths.ceil(min(tp[0, colx], tp[1, colx], tp[2, colx])), 0)
        max_tpx = min(maths.floor(max(tp[0, colx], tp[1, colx], tp[2, colx])), grid_nxyz[colx] - 1)
        if max_tpx < min_tpx:
            continue  # skip: triangle outside of grid
        min_tpy = max(maths.ceil(min(tp[0, coly], tp[1, coly], tp[2, coly])), 0)
        max_tpy = min(maths.floor(max(tp[0, coly], tp[1, coly], tp[2, coly])), grid_nxyz[coly] - 1)
        if max_tpy < min_tpy:
            continue  # skip: triangle outside of grid

        # 2. iterate over all points that fall within bounding box and
        # check whether points falls in triangle
        for py in range(min_tpy, max_tpy + 1, 1):
            for px in range(min_tpx, max_tpx + 1, 1):

                inside = False
                # 2a. use cross-product to work out Barycentric weights
                # this could be made prettier by refactoring a device function
                w1_denom = (tp[1, coly] - tp[0, coly]) * (tp[2, colx] - tp[0, colx]) - (tp[1, colx] - tp[0, colx]) * (
                    tp[2, coly] - tp[0, coly])
                w2_denom = tp[2, coly] - tp[0, coly]
                if w1_denom == 0.0 or w2_denom == 0.0:
                    inside = True  # point lies on a triangle which is actually a line (normally at boundaries)
                else:
                    w1 = (tp[0, colx] - numba.float64(px)) * (tp[2, coly] - tp[0, coly]) + (
                        numba.float64(py) - tp[0, coly]) * (tp[2, colx] - tp[0, colx])
                    w1 /= w1_denom
                    w2 = (numba.float64(py) - tp[0, coly] - w1 * (tp[1, coly] - tp[0, coly]))
                    w2 /= w2_denom
                    if w1 >= 0.0 and w2 >= 0.0 and (w1 + w2) <= 1.0:  # inside
                        inside = True  # point lies in triangle

                # 2b. the point is inside if Barycentric weights meet this condition
                if inside:
                    # 3. find intersection point with column centre
                    # 3a. Line start point in 3D which had a projection hit
                    line_p[axis] = numba.float64(grid_dxyz[axis]) / 2.0
                    line_p[2 - index1] = (py + 0.5) * grid_dxyz[2 - index1]  # kji / xyz & py=d1
                    line_p[2 - index2] = (px + 0.5) * grid_dxyz[2 - index2]  # kji / xyz & px=d2

                    # 3b. Line end point in 3D
                    for dim in range(3):
                        line_v[dim] = line_p[dim]
                    line_v[axis] = numba.float64(grid_dxyz[axis]) * (n_axis - numba.float64(0.5))  #!
                    for dim in range(3):  # for d in dimensions
                        line_v[dim] -= line_p[dim]

                    # 3c.find depth intersection
                    for dim in range(3):  # for d in dimensions
                        p01[dim] = (tp[1, dim] - tp[0, dim]) * grid_dxyz[dim]
                        p02[dim] = (tp[2, dim] - tp[0, dim]) * grid_dxyz[dim]

                    _cross_d(p01, p02, norm)  # normal to plane
                    _negative_d(line_v, line_rv)
                    _dot_d(line_rv, norm, denom)

                    if denom[0] == 0.0:
                        continue  # line is parallel to plane

                    for dim in range(3):
                        lp_t0[dim] = line_p[dim] - (tp[0, dim] + 0.5) * grid_dxyz[dim]

                    _dot_d(norm, lp_t0, t)
                    t[0] /= denom[0]
                    if t[0] < 0.0 - l_tol or t[0] > 1.0 + l_tol:
                        continue

                    _cross_d(p02, line_rv, tmp)
                    _dot_d(tmp, lp_t0, u)
                    u[0] /= denom[0]
                    if u[0] < 0.0 - t_tol or u[0] > 1.0 + t_tol:
                        continue

                    _cross_d(line_rv, p01, tmp)
                    _dot_d(tmp, lp_t0, v)
                    v[0] /= denom[0]
                    if v[0] < 0.0 - t_tol or u[0] + v[0] > 1.0 + t_tol:
                        continue

                    for dim in range(3):  # for d in dimensions
                        xyz[dim] = line_p[dim] + t[0] * line_v[dim]

                    # 4. mark the face corresponding to the grid and surface intersection at this point.
                    face = numba.int32((xyz[axis] - line_p[axis]) / grid_dxyz[axis])

                    if face == -1:  # handle rounding precision issues
                        face = 0
                    elif face == n_faces:
                        face -= 1
                    assert 0 <= face < n_faces

                    face_idx[index1] = int(py)
                    face_idx[index2] = int(px)
                    face_idx[2 - axis] = int(face)

                    faces[face_idx[0], face_idx[1], face_idx[2]] = True

                    if return_depths:
                        depths[face_idx[0], face_idx[1], face_idx[2]] = xyz[2]

                    if return_offsets:
                        offsets[face_idx[0], face_idx[1], face_idx[2]] = xyz[axis] - ((face + 1) * grid_dxyz[axis])

                    if return_normal_vectors:
                        for dim in range(3):
                            line_p[dim] = (tp[0, dim] - tp[1, dim]) * grid_dxyz[dim]
                            line_v[dim] = (tp[0, dim] - tp[2, dim]) * grid_dxyz[dim]
                        _cross_d(line_p, line_v, tmp)
                        _norm_d(tmp, v)
                        for dim in range(3):
                            normals[face_idx[0], face_idx[1], face_idx[2], dim] = (-1.0 * tmp[dim] / v[0])
                        norm_idx[index2] = int(px)
                        if normals[norm_idx[0], norm_idx[1], norm_idx[2], 2] > 0.0:
                            for dim in range(3):
                                normals[face_idx[0], face_idx[1], face_idx[2], dim] *= -1.0

                    if return_triangles:
                        triangle_per_face[face_idx[0], face_idx[1], face_idx[2]] = triangle_num


@cuda.jit
def _diffuse_closed_faces(a, k_faces, j_faces, i_faces, index1, index2, axis, start, stop, inc):

    tidx, tidy = cuda.grid(2)
    stridex, stridey = cuda.gridsize(2)
    maxidx, maxidy = a.shape[index1] - 2, a.shape[index2] - 2
    indices = numba.cuda.local.array(3, numba.int32)
    for D1 in range(tidx, maxidx, stridex):  # k vectorized
        for D2 in range(tidy, maxidy, stridey):  # j vectorized
            indices[index1] = D1 + 1
            indices[index2] = D2 + 1
            for D3 in range(start, stop, inc):  # iterate along i in kj-planes
                indices[axis] = D3 + 1
                i, j, k = indices[:]
                iF, jF, kF = i - 1, j - 1, k - 1  # faces arrays aren't padded
                fault_above = k_faces[iF - 1, jF, kF]
                fault_below = k_faces[iF, jF, kF]
                fault_left = j_faces[iF, jF - 1, kF]
                fault_right = j_faces[iF, jF, kF]
                fault_behind = i_faces[iF, jF, kF - 1]
                fault_back = i_faces[iF, jF, kF]
                cuda.syncthreads()
                a[i, j, k] = ((a[i - 1, j, k] and (not fault_above)) or (a[i + 1, j, k] and (not fault_below)) or
                              (a[i, j - 1, k] and (not fault_left)) or (a[i, j + 1, k] and (not fault_right)) or
                              (a[i, j, k - 1] and
                               (not fault_behind)) or (a[i, j, k + 1] and
                                                       (not fault_back)) or a[i, j, k])  # already closed
                cuda.syncthreads()


def bisector_from_faces_cuda(
    grid_extent_kji: Tuple[int, int, int],
    k_faces: np.ndarray,
    j_faces: np.ndarray,
    i_faces: np.ndarray,
) -> Tuple[np.ndarray, bool]:
    """Returns a numpy bool array denoting the bisection of the grid by the face sets, using GPUs.

    arguments:
        grid_extent_kji (triple int): the shape of the grid
        k_faces, j_faces, i_faces (numpy bool arrays): True where an internal grid face forms part of the
            bisecting surface

    returns:
        (numpy bool array of shape grid_extent_kji, bool) where the array is set True for cells on one side
        of the face sets deemed to be shallower (more strictly, lower K index on average); set False for cells
        on othe side; the bool value is True if the surface is a curtain (vertical), otherwise False

    notes:
        the face sets must form a single 'sealed' cut of the grid (eg. not waving in and out of the grid);
        any 'boxed in' parts of the grid (completely enclosed by bisecting faces) will be consistently
        assigned to either the True or False part
    """
    assert len(grid_extent_kji) == 3
    padded_extent_kji = (
        grid_extent_kji[0] + 2,
        grid_extent_kji[1] + 2,
        grid_extent_kji[2] + 2,
    )
    a = cupy.zeros(padded_extent_kji, dtype = bool)
    a[1, 1, 1] = True

    a_count = a_count_before = 0
    blockSize = (16, 16)
    gridSize_k = (
        (grid_extent_kji[1] + blockSize[0] - 1) // blockSize[0],
        (grid_extent_kji[2] + blockSize[1] - 1) // blockSize[1],
    )
    gridSize_j = (
        (grid_extent_kji[0] + blockSize[0] - 1) // blockSize[0],
        (grid_extent_kji[2] + blockSize[1] - 1) // blockSize[1],
    )
    gridSize_i = (
        (grid_extent_kji[0] + blockSize[0] - 1) // blockSize[1],
        (grid_extent_kji[1] + blockSize[1] - 1) // blockSize[1],
    )

    while True:
        # forward sweeps
        _diffuse_closed_faces[gridSize_k, blockSize](a, k_faces, j_faces, i_faces, 1, 2, 0, 0, grid_extent_kji[0],
                                                     1)  # k-direction
        _diffuse_closed_faces[gridSize_j, blockSize](a, k_faces, j_faces, i_faces, 0, 2, 1, 0, grid_extent_kji[1],
                                                     1)  # j-direction
        _diffuse_closed_faces[gridSize_i, blockSize](a, k_faces, j_faces, i_faces, 0, 1, 2, 0, grid_extent_kji[2],
                                                     1)  # i-direction
        # reverse sweeps
        _diffuse_closed_faces[gridSize_k, blockSize](a, k_faces, j_faces, i_faces, 1, 2, 0, grid_extent_kji[0] - 1, -1,
                                                     -1)  # k-direction
        _diffuse_closed_faces[gridSize_j, blockSize](a, k_faces, j_faces, i_faces, 0, 2, 1, grid_extent_kji[1] - 1, -1,
                                                     -1)  # j-direction
        _diffuse_closed_faces[gridSize_i, blockSize](a, k_faces, j_faces, i_faces, 0, 1, 2, grid_extent_kji[2] - 1, -1,
                                                     -1)  # i-direction

        a_count = cupy.count_nonzero(a)

        if a_count == a_count_before:
            break
        a_count_before = a_count

    a = cupy.asnumpy(a[1:-1, 1:-1, 1:-1])
    cell_count = a.size
    assert (1 <= a_count < cell_count), "face set for surface is leaky or empty (surface does not intersect grid)"

    # find mean K for a cells and not a cells; if not a cells mean K is lesser (ie shallower), negate a
    layer_cell_count = grid_extent_kji[1] * grid_extent_kji[2]
    a_k_sum = 0
    not_a_k_sum = 0
    for k in range(grid_extent_kji[0]):
        a_layer_count = np.count_nonzero(a[k])
        a_k_sum += (k + 1) * a_layer_count
        not_a_k_sum += (k + 1) * (layer_cell_count - a_layer_count)
    a_mean_k = float(a_k_sum) / float(a_count)
    not_a_mean_k = float(not_a_k_sum) / float(cell_count - a_count)
    is_curtain = False
    if a_mean_k > not_a_mean_k:
        a[:] = np.logical_not(a)
    elif abs(a_mean_k - not_a_mean_k) <= 0.01:
        # log.warning('unable to determine which side of surface is shallower')
        is_curtain = True

    return a, is_curtain


def find_faces_to_represent_surface_regular_cuda_sgpu(
    grid,
    surfaces,
    name,
    title = None,
    agitate = False,
    feature_type = "fault",
    progress_fn = None,
    return_properties = None,
    i_surface = 0,
    i_gpu = 0,
    gcs_list = None,
    props_dict_list = None,
):
    """Returns a grid connection set containing those cell faces which are deemed to represent the surface, using GPUs.

    arguments:
        grid (RegularGrid): the grid for which to create a grid connection set representation of the surface;
           must be aligned, ie. I with +x, J with +y, K with +z and local origin of (0.0, 0.0, 0.0)
        surface (list(Surface)): the surface to be intersected with the grid
        name (str): the feature name to use in the grid connection set
        title (str, optional): the citation title to use for the grid connection set; defaults to name
        agitate (bool, default False): if True, the points of the surface are perturbed by a small random
           offset, which can help if the surface has been built from a regular mesh with a periodic resonance
           with the grid
        feature_type (str, default 'fault'): 'fault', 'horizon' or 'geobody boundary'
        progress_fn (f(x: float), optional): a callback function to be called at intervals by this function;
           the argument will progress from 0.0 to 1.0 in unspecified and uneven increments

    returns:
        gcs  or  (gcs, gcs_props)
        where gcs is a new GridConnectionSet with a single feature, not yet written to hdf5 nor xml created;
        gcs_props is a dictionary mapping from requested return_properties string to numpy array

    notes:
        this function is designed for aligned regular grids only;
        this function can handle the surface and grid being in different coordinate reference systems, as
        long as the implicit parent crs is shared;
        no trimming of the surface is carried out here: for computational efficiency, it is recommended
        to trim first;
        organisational objects for the feature are created if needed
    """
    # todo: update with extra arguments to keep functionality aligned with find_faces...regular_optimised

    cuda.select_device(i_gpu)  # bind device to thread
    device = (cuda.get_current_device())  # if no GPU present - this will throw an exception and fall back to CPU

    assert isinstance(grid, grr.RegularGrid)
    assert grid.is_aligned
    return_triangles = False
    return_normal_vectors = False
    return_depths = False
    return_offsets = False
    return_bisector = False
    return_flange_bool = False
    if return_properties:
        assert all([
            p in [
                "triangle",
                "depth",
                "offset",
                "normal vector",
                "grid bisector",
                "flange bool",
            ] for p in return_properties
        ])
        return_triangles = "triangle" in return_properties
        return_normal_vectors = "normal vector" in return_properties
        return_depths = "depth" in return_properties
        return_offsets = "offset" in return_properties
        return_bisector = "grid bisector" in return_properties
        return_flange_bool = "flange bool" in return_properties
        if return_flange_bool:
            return_triangles = True

    if title is None:
        title = name

    if progress_fn is not None:
        progress_fn(0.0)

    # prepare surfaces
    surface = surfaces[i_surface]  # get surface under consideration
    log.debug(f"intersecting surface {surface.title} with regular grid {grid.title} on a GPU")
    # log.debug(f'grid extent kji: {grid.extent_kji}')

    # print some information about the CUDA card
    log.debug(f"{device.name} | Device Controller {i_gpu} | " +
              f"CC {device.COMPUTE_CAPABILITY_MAJOR}.{device.COMPUTE_CAPABILITY_MINOR} | " +
              f"Processing surface {i_surface}")
    # get device attributes to calculate thread dimensions
    nSMs = device.MULTIPROCESSOR_COUNT  # number of SMs
    maxBlockSize = (device.MAX_BLOCK_DIM_X / 2)  # max number of threads per block in x-dim
    gridSize = 2 * nSMs  # prefer 2*nSMs blocks for full occupancy
    # take the reverse diagonal for relationship between xyz & ijk
    grid_dxyz = (
        grid.block_dxyz_dkji[2, 0],
        grid.block_dxyz_dkji[1, 1],
        grid.block_dxyz_dkji[0, 2],
    )
    # extract polygons from surface
    with compiler_lock:  # HDF5 handles seem not to be threadsafe
        triangles, points = surface.triangles_and_points()
    assert (triangles is not None and points is not None), f"surface {surface.title} is empty"

    if agitate:
        points += 1.0e-5 * (np.random.random(points.shape) - 0.5)  # +/- uniform err.
    # log.debug(f'surface: {surface.title}; p0: {points[0]}; crs uuid: {surface.crs_uuid}')
    # log.debug(f'surface min xyz: {np.min(points, axis = 0)}')
    # log.debug(f'surface max xyz: {np.max(points, axis = 0)}')
    if not bu.matching_uuids(grid.crs_uuid, surface.crs_uuid):
        log.debug("converting from surface crs to grid crs")
        s_crs = rqc.Crs(surface.model, uuid = surface.crs_uuid)
        s_crs.convert_array_to(grid.crs, points)
        surface.crs_uuid = grid.crs.uuid
        # log.debug(f'surface: {surface.title}; p0: {points[0]}; crs uuid: {surface.crs_uuid}')
        # log.debug(f'surface min xyz: {np.min(points, axis = 0)}')
        # log.debug(f'surface max xyz: {np.max(points, axis = 0)}')

    p_tri_xyz = points[triangles]
    p_tri_xyz_d = cupy.asarray(p_tri_xyz)

    # K direction (xy projection)
    if grid.nk > 1:
        log.debug("searching for k faces")
        k_faces = np.zeros((grid.nk - 1, grid.nj, grid.ni), dtype = bool)
        k_triangles = (np.full((grid.nk - 1, grid.nj, grid.ni), -1, dtype = int) if return_triangles else np.full(
            (1, 1, 1), -1, dtype = int))
        k_depths = (np.full((grid.nk - 1, grid.nj, grid.ni), np.nan) if return_depths else np.full((1, 1, 1), np.nan))
        k_offsets = (np.full((grid.nk - 1, grid.nj, grid.ni), np.nan) if return_offsets else np.full((1, 1, 1), np.nan))
        k_normals = (np.full((grid.nk - 1, grid.nj, grid.ni, 3), np.nan) if return_normal_vectors else np.full(
            (1, 1, 1, 1), np.nan))
        k_faces_d = cupy.asarray(k_faces)
        k_triangles_d = cupy.asarray(k_triangles)
        k_depths_d = cupy.asarray(k_depths)
        k_offsets_d = cupy.asarray(k_offsets)
        k_normals_d = cupy.asarray(k_normals)
        colx = 0
        coly = 1
        axis = 2
        index1 = 1
        index2 = 2
        blockSize = ((p_tri_xyz.shape[0] - 1) // (gridSize - 1) if
                     (p_tri_xyz.shape[0] < gridSize * maxBlockSize) else 64)  # prefer factors of 32 (threads per warp)
        log.debug(
            f"Executing polygon-intersection GPU-kernel along k-axis using gridSize={gridSize}, blockSize={blockSize}")
        project_polygons_to_surfaces[gridSize, blockSize](
            k_faces_d,
            p_tri_xyz_d,
            axis,
            index1,
            index2,
            colx,
            coly,
            grid.ni,
            grid.nj,
            grid.nk,
            grid_dxyz[0],
            grid_dxyz[1],
            grid_dxyz[2],
            0.0,
            0.0,
            return_normal_vectors,
            k_normals_d,
            return_depths,
            k_depths_d,
            return_offsets,
            k_offsets_d,
            return_triangles,
            k_triangles_d,
        )
        k_faces = cupy.asnumpy(k_faces_d)
        if not return_bisector:
            del k_faces_d
        k_triangles = cupy.asnumpy(k_triangles_d)
        del k_triangles_d
        k_depths = cupy.asnumpy(k_depths_d)
        del k_depths_d
        k_offsets = cupy.asnumpy(k_offsets_d)
        del k_offsets_d
        k_normals = cupy.asnumpy(k_normals_d)
        del k_normals_d
        log.debug(f"k face count: {np.count_nonzero(k_faces)}")
    else:
        k_faces = None

    if progress_fn is not None:
        progress_fn(0.3)

    # J direction (xz projection)
    if grid.nj > 1:
        log.debug("searching for j faces")
        j_faces = np.zeros((grid.nk, grid.nj - 1, grid.ni), dtype = bool)
        j_triangles = (np.full((grid.nk, grid.nj - 1, grid.ni), -1, dtype = int) if return_triangles else np.full(
            (1, 1, 1), -1, dtype = int))
        j_depths = (np.full((grid.nk, grid.nj - 1, grid.ni), np.nan) if return_depths else np.full((1, 1, 1), np.nan))
        j_offsets = (np.full((grid.nk, grid.nj - 1, grid.ni), np.nan) if return_offsets else np.full((1, 1, 1), np.nan))
        j_normals = (np.full((grid.nk, grid.nj - 1, grid.ni, 3), np.nan) if return_normal_vectors else np.full(
            (1, 1, 1, 1), np.nan))
        j_faces_d = cupy.asarray(j_faces)
        j_triangles_d = cupy.asarray(j_triangles)
        j_depths_d = cupy.asarray(j_depths)
        j_offsets_d = cupy.asarray(j_offsets)
        j_normals_d = cupy.asarray(j_normals)
        colx = 0
        coly = 2
        axis = 1
        index1 = 0
        index2 = 2
        blockSize = ((p_tri_xyz.shape[0] - 1) // (gridSize - 1) if
                     (p_tri_xyz.shape[0] < gridSize * maxBlockSize) else 64)  # prefer factors of 32 (threads per warp)
        log.debug(
            f"Executing polygon-intersection GPU-kernel along j-axis using gridSize={gridSize}, blockSize={blockSize}")
        project_polygons_to_surfaces[gridSize, blockSize](
            j_faces_d,
            p_tri_xyz_d,
            axis,
            index1,
            index2,
            colx,
            coly,
            grid.ni,
            grid.nj,
            grid.nk,
            grid_dxyz[0],
            grid_dxyz[1],
            grid_dxyz[2],
            0.0,
            0.0,
            return_normal_vectors,
            j_normals_d,
            return_depths,
            j_depths_d,
            return_offsets,
            j_offsets_d,
            return_triangles,
            j_triangles_d,
        )
        j_faces = cupy.asnumpy(j_faces_d)
        if not return_bisector:
            del j_faces_d
        j_triangles = cupy.asnumpy(j_triangles_d)
        del j_triangles_d
        j_depths = cupy.asnumpy(j_depths_d)
        del j_depths_d
        j_offsets = cupy.asnumpy(j_offsets_d)
        del j_offsets_d
        j_normals = cupy.asnumpy(j_normals_d)
        del j_normals_d
        log.debug(f"j face count: {np.count_nonzero(j_faces)}")
    else:
        j_faces = None

    if progress_fn is not None:
        progress_fn(0.6)

    # I direction (yz projection)
    if grid.ni > 1:
        log.debug("searching for i faces")
        i_faces = np.zeros((grid.nk, grid.nj, grid.ni - 1), dtype = bool)
        i_triangles = (np.full((grid.nk, grid.nj, grid.ni - 1), -1, dtype = int) if return_triangles else np.full(
            (1, 1, 1), -1, dtype = int))
        i_depths = (np.full((grid.nk, grid.nj, grid.ni - 1), np.nan) if return_depths else np.full((1, 1, 1), np.nan))
        i_offsets = (np.full((grid.nk, grid.nj, grid.ni - 1), np.nan) if return_offsets else np.full((1, 1, 1), np.nan))
        i_normals = (np.full((grid.nk, grid.nj, grid.ni - 1, 3), np.nan) if return_normal_vectors else np.full(
            (1, 1, 1, 1), np.nan))
        i_faces_d = cupy.asarray(i_faces)
        i_triangles_d = cupy.asarray(i_triangles)
        i_depths_d = cupy.asarray(i_depths)
        i_offsets_d = cupy.asarray(i_offsets)
        i_normals_d = cupy.asarray(i_normals)
        colx = 1
        coly = 2
        axis = 0
        index1 = 0
        index2 = 1
        blockSize = ((p_tri_xyz.shape[0] - 1) // (gridSize - 1) if
                     (p_tri_xyz.shape[0] < gridSize * maxBlockSize) else 64)  # prefer factors of 32 (threads per warp)
        log.debug(
            f"Executing polygon-intersection GPU-kernel along i-axis using gridSize={gridSize}, blockSize={blockSize}")
        project_polygons_to_surfaces[gridSize, blockSize](
            i_faces_d,
            p_tri_xyz_d,
            axis,
            index1,
            index2,
            colx,
            coly,
            grid.ni,
            grid.nj,
            grid.nk,
            grid_dxyz[0],
            grid_dxyz[1],
            grid_dxyz[2],
            0.0,
            0.0,
            return_normal_vectors,
            i_normals_d,
            return_depths,
            i_depths_d,
            return_offsets,
            i_offsets_d,
            return_triangles,
            i_triangles_d,
        )
        i_faces = cupy.asnumpy(i_faces_d)
        if not return_bisector:
            del i_faces_d
        i_triangles = cupy.asnumpy(i_triangles_d)
        del i_triangles_d
        i_depths = cupy.asnumpy(i_depths_d)
        del i_depths_d
        i_offsets = cupy.asnumpy(i_offsets_d)
        del i_offsets_d
        i_normals = cupy.asnumpy(i_normals_d)
        del i_normals_d
        log.debug(f"i face count: {np.count_nonzero(i_faces)}")
    else:
        i_faces = None

    del p_tri_xyz_d

    if progress_fn is not None:
        progress_fn(0.9)

    log.debug("converting face sets into grid connection set")
    gcs_list[i_surface] = rqf.GridConnectionSet(
        grid.model,
        grid = grid,
        k_faces = k_faces,
        j_faces = j_faces,
        i_faces = i_faces,
        k_sides = None,
        j_sides = None,
        i_sides = None,
        feature_name = name,
        feature_type = feature_type,
        title = title,
        create_organizing_objects_where_needed = True,
    )

    # NB. following assumes faces have been added to gcs in a particular order!
    if return_triangles:
        k_tri_list = (np.empty((0,)) if k_triangles is None else k_triangles[rgs_ff._where_true(k_faces)])
        j_tri_list = (np.empty((0,)) if j_triangles is None else j_triangles[rgs_ff._where_true(j_faces)])
        i_tri_list = (np.empty((0,)) if i_triangles is None else i_triangles[rgs_ff._where_true(i_faces)])
        all_tris = np.concatenate((k_tri_list, j_tri_list, i_tri_list), axis = 0)
        # log.debug(f'gcs count: {gcs.count}; all triangles shape: {all_tris.shape}')
        assert all_tris.shape == (gcs_list[i_surface].count,)

    # NB. following assumes faces have been added to gcs in a particular order!
    if return_depths:
        k_depths_list = (np.empty((0,)) if k_depths is None else k_depths[rgs_ff._where_true(k_faces)])
        j_depths_list = (np.empty((0,)) if j_depths is None else j_depths[rgs_ff._where_true(j_faces)])
        i_depths_list = (np.empty((0,)) if i_depths is None else i_depths[rgs_ff._where_true(i_faces)])
        all_depths = np.concatenate((k_depths_list, j_depths_list, i_depths_list), axis = 0)
        # log.debug(f'gcs count: {gcs.count}; all depths shape: {all_depths.shape}')
        assert all_depths.shape == (gcs_list[i_surface].count,)

    # NB. following assumes faces have been added to gcs in a particular order!
    if return_offsets:
        k_offsets_list = (np.empty((0,)) if k_offsets is None else k_offsets[rgs_ff._where_true(k_faces)])
        j_offsets_list = (np.empty((0,)) if j_offsets is None else j_offsets[rgs_ff._where_true(j_faces)])
        i_offsets_list = (np.empty((0,)) if i_offsets is None else i_offsets[rgs_ff._where_true(i_faces)])
        all_offsets = np.concatenate((k_offsets_list, j_offsets_list, i_offsets_list), axis = 0)
        # log.debug(f'gcs count: {gcs.count}; all offsets shape: {all_offsets.shape}')
        assert all_offsets.shape == (gcs_list[i_surface].count,)

    if return_flange_bool:
        flange_bool_uuid = surface.model.uuid(title = "flange bool",
                                              obj_type = "DiscreteProperty",
                                              related_uuid = surface.uuid)
        assert (flange_bool_uuid is not None), f"No flange bool property found for surface: {surface.title}"
        flange_bool = rqp.Property(surface.model, uuid = flange_bool_uuid)
        flange_array = flange_bool.array_ref()
        all_flange = np.take(flange_array, all_tris)
        assert all_flange.shape == (gcs_list[i_surface].count,)

    # NB. following assumes faces have been added to gcs in a particular order!
    if return_normal_vectors:
        k_normals_list = (np.empty((0, 3)) if k_normals is None else k_normals[rgs_ff._where_true(k_faces)])
        j_normals_list = (np.empty((0, 3)) if j_normals is None else j_normals[rgs_ff._where_true(j_faces)])
        i_normals_list = (np.empty((0, 3)) if i_normals is None else i_normals[rgs_ff._where_true(i_faces)])
        all_normals = np.concatenate((k_normals_list, j_normals_list, i_normals_list), axis = 0)
        # log.debug(f'gcs count: {gcs.count}; all normals shape: {all_normals.shape}')
        assert all_normals.shape == (gcs_list[i_surface].count, 3)

    # note: following is a grid cells property, not a gcs property
    if return_bisector:
        bisector, is_curtain = bisector_from_faces_cuda(tuple(grid.extent_kji), k_faces_d, j_faces_d, i_faces_d)
        del k_faces_d, j_faces_d, i_faces_d

    if progress_fn is not None:
        progress_fn(1.0)

    # if returning properties, construct dictionary
    if return_properties:
        props_dict_list[i_surface] = {}
        if return_triangles:
            props_dict_list[i_surface]["triangle"] = all_tris
        if return_depths:
            props_dict_list[i_surface]["depth"] = all_depths
        if return_offsets:
            props_dict_list[i_surface]["offset"] = all_offsets
        if return_normal_vectors:
            props_dict_list[i_surface]["normal vector"] = all_normals
        if return_bisector:
            props_dict_list[i_surface]["grid bisector"] = (bisector, is_curtain)
        if return_flange_bool:
            props_dict_list[i_surface]["flange bool"] = all_flange


def find_faces_to_represent_surface_regular_cuda_mgpu(
    grid,
    surface,
    name,
    title = None,
    agitate = False,
    feature_type = "fault",
    progress_fn = None,
    return_properties = None,
):
    """Returns a grid connection set containing those cell faces which are deemed to represent the surface, using GPUs.

    arguments:
        grid (RegularGrid): the grid for which to create a grid connection set representation of the surface;
           must be aligned, ie. I with +x, J with +y, K with +z and local origin of (0.0, 0.0, 0.0)
        surface (Surface or list(Surface)): the surface(s) to be intersected with the grid
        name (str): the feature name to use in the grid connection set
        title (str, optional): the citation title to use for the grid connection set; defaults to name
        agitate (bool, default False): if True, the points of the surface are perturbed by a small random
           offset, which can help if the surface has been built from a regular mesh with a periodic resonance
           with the grid
        feature_type (str, default 'fault'): 'fault', 'horizon' or 'geobody boundary'
        progress_fn (f(x: float), optional): a callback function to be called at intervals by this function;
           the argument will progress from 0.0 to 1.0 in unspecified and uneven increments

    returns:
        gcs  or  (gcs, gcs_props)
        where gcs is a new GridConnectionSet with a single feature, not yet written to hdf5 nor xml created;
        gcs_props is a dictionary mapping from requested return_properties string to numpy array

    notes:
        this function is designed for aligned regular grids only;
        this function can handle the surface and grid being in different coordinate reference systems, as
        long as the implicit parent crs is shared;
        no trimming of the surface is carried out here: for computational efficiency, it is recommended
        to trim first;
        organisational objects for the feature are created if needed
    """

    surfaces = surface if isinstance(surface, list) else [surface]

    n_surfs = len(surfaces)
    n_gpus = len(cuda.list_devices())
    log.debug("distributing %d surface between %d GPUs" % (n_surfs, n_gpus))
    gcs_list = [None] * n_surfs
    props_dict_list = [None] * n_surfs
    threads = [None] * n_gpus
    for i_surface in range(n_surfs):
        threads[i_surface % n_gpus] = threading.Thread(
            target = find_faces_to_represent_surface_regular_cuda_sgpu,
            args = (
                grid,
                surfaces,
                name,
                title,
                agitate,
                feature_type,
                progress_fn,
                return_properties,
                i_surface,
                i_surface % n_gpus,
                gcs_list,
                props_dict_list,
            ),
        )
        threads[i_surface % n_gpus].start()  # start parallel run
        # if this is the last GPU available or we're at the last array ...
        if (i_surface + 1) % n_gpus == 0 or (i_surface + 1) == n_surfs:
            # ... sync all the GPUs being used
            for i_gpu in range(i_surface % n_gpus + 1):  # up to the number of GPUs being used
                threads[i_gpu].join()  # rejoin the main thread (syncthreads)

    if n_surfs > 1:
        return (gcs_list, props_dict_list) if return_properties else gcs_list
    else:
        return (gcs_list[0], props_dict_list[0]) if return_properties else gcs_list[0]



"""Class for a welllog, representing resqml properties for well logs"""

import logging

log = logging.getLogger(__name__)

import resqpy.property
import resqpy.property.property_collection as rqp_pc


class WellLog:
    """Thin wrapper class around RESQML properties for well logs."""

    def __init__(self, collection, uuid):
        """Create a well log from a part name."""

        self.collection: rqp_pc.PropertyCollection = collection
        self.model = collection.model
        self.uuid = uuid

        part = self.model.part_for_uuid(uuid)
        indexable = self.collection.indexable_for_part(part)
        if indexable != 'nodes':
            raise NotImplementedError('well frame related property does not have nodes as indexable element')

        #: Name of log
        self.title = self.model.citation_title_for_part(part)

        #: Unit of measure
        self.uom = self.collection.uom_for_part(part)

    def values(self):
        """Return log data as numpy array.

        Note:
           may return 2D numpy array with shape (num_depths, num_columns).
        """

        part = self.model.part_for_uuid(self.uuid)
        return self.collection.cached_part_array_ref(part)



"""Class for RESQML Structural Organization Interpretation organizational objects."""

from resqpy.olio.base import BaseResqpy

valid_ordering_criteria = ('age', 'apparent depth', 'measured depth')

valid_contact_relationships = ("frontier feature to frontier feature", "genetic boundary to frontier feature",
                               "genetic boundary to genetic boundary", "genetic boundary to tectonic boundary",
                               "stratigraphic unit to frontier feature", "stratigraphic unit to stratigraphic unit",
                               "tectonic boundary to frontier feature", "tectonic boundary to genetic boundary",
                               "tectonic boundary to tectonic boundary")


class StructuralOrganizationInterpretation(BaseResqpy):
    """Class for RESQML Structural Organization Interpretation organizational objects."""

    resqml_type = 'StructuralOrganizationInterpretation'

    def __init__(self):
        """Initialise a structural organisation interpretation."""
        self.ordering_criterion = 'age'  #: one of 'age' (youngest to oldest!), 'apparent depth', or 'measured depth'
        self.fault_uuid_list = None  #: list of uuids of fault interpretation which intersect the structural object
        self.horizon_tuple_list = None  #: list of horizon interpretation uuids along with index and rank
        self.sides_uuid_list = None  #: list of uuids of interpretation objects for sides of structural object
        self.top_frontier_uuid_list = None  #: list of uuids of interpretation objects for top of structural object
        self.bottom_frontier_uuid_list = None  #: list of uuids of interpretation objects for base of structural object
        self.contact_interpretations = None  #: list of contact interpretations

    # TODO



"""TriMesh class using equilateral triangles, derived from Mesh."""

import logging

log = logging.getLogger(__name__)

import math as maths
import numpy as np

import resqpy.surface as rqs
import resqpy.olio.vector_utilities as vec

root_3 = maths.sqrt(3.0)
root_3_by_2 = root_3 / 2.0


class TriMesh(rqs.Mesh):
    """Class of mesh using equilateral triangles in the xy plane."""

    def __init__(self,
                 parent_model,
                 uuid = None,
                 t_side = None,
                 nj = None,
                 ni = None,
                 origin = None,
                 z_values = None,
                 z_uom = None,
                 surface_role = 'map',
                 crs_uuid = None,
                 title = None,
                 originator = None,
                 extra_metadata = None):
        """Initialises a TriMesh object from xml, or from arguments.

        arguments:
           parent_model (model.Model object): the model to which this Mesh object will be associated
           uuid (uuid.UUID, optional): the uuid of an existing RESQML obj_Grid2dRepresentation object from which
               this resqpy Mesh object is populated
           t_side (float, optional): the length of a side of each triangle; units are xy units of crs;
               required if uuid is None
           nj (int, optional): the number of nodes (NB. not 'cells') in the j axis of the mesh;
               required if uuid is None
           ni (int, optional): the number of nodes in the i axis of the mesh; required if uuid is None
           origin (triple float, optional): the xyz origin of the regular mesh; z usually zero; defaults
               to triple zero
           z_values (numpy int array of shape (nj, ni), optional): z values; recommended if uuid is None or will
               default to zero
           z_uom (str, optional): recommended if uuid is None and z values are not in the crs z space
           surface_role (string, default 'map'): 'map' or 'pick'; ignored if uuid is not None
           crs_uuid (uuid.Uuid or string, optional): required if generating a regular mesh, the uuid of the crs
           title (str, optional): the citation title to use for a new mesh;
              ignored if uuid is not None
           originator (str, optional): the name of the person creating the mesh, defaults to login id;
              ignored if uuid is not None
           extra_metadata (dict, optional): string key, value pairs to add as extra metadata for the mesh;
              ignored if uuid is not None

        returns:
           the newly created TriMesh object

        note:
           using the z_uom argument will result in the tri mesh's full_array_ref() method returning data with
           x and y values in the object's crs, whilst z values will be in the given uom
        """

        if uuid is None:
            assert t_side > 0.0
            assert nj > 1 and ni > 1
            assert z_values is None or z_values.shape == (nj, ni)
            assert origin is None or len(origin) == 3
            assert crs_uuid is not None
            xyz = np.zeros((nj, ni, 3), dtype = float)
            xyz[:, :, 0] = np.expand_dims(np.arange(ni).astype(float), axis = 0)  # x
            xyz[1::2, :, 0] += 0.5  # offset every second row by half triangle side length
            xyz[:, :, 1] = root_3_by_2 * np.expand_dims(np.arange(nj).astype(float), axis = 1)
            xyz *= t_side
            if z_values is not None:
                xyz[:, :, 2] = z_values
            if origin is not None:
                xyz += np.expand_dims(np.expand_dims(origin, axis = 0), axis = 0)
            super().__init__(parent_model,
                             mesh_flavour = 'explicit',
                             xyz_values = xyz,
                             nj = nj,
                             ni = ni,
                             surface_role = surface_role,
                             crs_uuid = crs_uuid,
                             title = title,
                             originator = originator,
                             extra_metadata = extra_metadata)
            self.t_side = t_side
            self.origin = None
            if origin is not None:
                self.origin = np.array(origin, dtype = float)
                assert self.origin.shape == (3,)
            self.z_uom = z_uom
            self.extra_metadata['t side'] = str(t_side)
            if z_uom:
                self.extra_metadata['z uom'] = str(z_uom)
            self.represented_interpretation_root = None

        else:
            super().__init__(parent_model, uuid = uuid)
            assert self.flavour == 'explicit', 'mesh flavour must be explicit for TriMesh objects'
            assert hasattr(self, 'extra_metadata')
            t_side = self.extra_metadata.get('t side')
            assert t_side is not None, 'triangle side length missing in TriMesh extra metadata'
            self.t_side = float(t_side)
            self.z_uom = self.extra_metadata.get('z uom')
            origin = self.full_array_ref()[0, 0]
            if np.all(np.isclose(origin, 0.0)):
                self.origin = None
            else:
                self.origin = origin

    @classmethod
    def from_tri_mesh_and_z_values(cls, tri_mesh, z_values, z_uom = None, title = None, surface_role = 'map'):
        """Create a new tri mesh with the same xy pattern as an existing one, updating z."""
        assert z_values.shape == (tri_mesh.nj, tri_mesh.ni)
        if not z_uom:
            z_uom = tri_mesh.z_uom
        if not title:
            title = tri_mesh.title
        tm = cls(tri_mesh.model,
                 uuid = None,
                 t_side = tri_mesh.t_side,
                 nj = tri_mesh.nj,
                 ni = tri_mesh.ni,
                 origin = tri_mesh.origin,
                 z_values = z_values,
                 z_uom = z_uom,
                 surface_role = surface_role,
                 crs_uuid = tri_mesh.crs_uuid,
                 title = title)
        return tm

    def tji_for_xy(self, xy):
        """Return (tj, ti) indices of triangle containing point xy (x, y).

        note:
            tj triangle indices are in the range 0..nj - 2 inclusive
            ti triangle indices are in the range 0..(ni - 2) * 2 + 1 inclusive
        """
        tji, _ = self.tji_tc_for_xy(xy)
        return tji

    def tji_for_xy_array(self, xy_array):
        """Return array of (tj, ti) indices of triangles containing points xy_array (..., 2 or 3).

        note:
            tj triangle indices are in the range 0..nj - 2 inclusive
            ti triangle indices are in the range 0..(ni - 2) * 2 + 1 inclusive
        """
        tji, _ = self.tji_tc_for_xy_array(xy_array)
        return tji

    def tji_tc_for_xy(self, xy):
        """Return indices of triangle containing point xy (x, y), and trilinear coordinates within triangle.

        arguments:
            xy (pair of floats): coordinates x, y of a point of interest

        returns:
            ((tj, ti), (f0, f1, f2)): the indices of the triangle containing the point xy, and the trilinear
            coordinates of the point within the triangle; f2 is component from base edge towards point 2;
            f1 is component towards point 1; f0 is component towards point 0; the trilinear coordinates
            sum to one and can be used as weights to interpolate z values at point xy
        """
        x, y = xy
        if self.origin is not None:
            x -= self.origin[0]
            y -= self.origin[1]
        jp = y / (self.t_side * root_3_by_2)
        j = maths.floor(jp)
        if not 0 <= j < self.nj - 1:
            return (None, None)
        if j % 2:
            fy = float(j + 1) - jp
        else:
            fy = jp - float(j)
        ip = x / self.t_side - 0.5 * fy
        i = maths.floor(ip)
        if not 0 <= i < self.ni - 1:
            return (None, None)
        fx = ip - float(i)
        i *= 2
        if fx > 1.0 - fy:
            i += 1
            fx -= 1.0 - fy
            fy = 1.0 - fy
        return ((j, i), (1.0 - (fx + fy), fx, fy))

    def tji_tc_for_xy_array(self, xy_array):
        """Return indices of triangles containing array of points xy, and trilinear coordinates within triangles.

        arguments:
            xy_array (numpy float array of shape (..., 2 or 3)): coordinates x, y of points of interest

        returns:
            pair of numpy arrays of shape (..., 2) and (..., 3); the first has int elements, being
            (tj, ti) indices of the triangles containing each point xy; the second has float elements,
            being the trilinear coordinates of the points within the triangles; f2 is component from
            base edge towards point 2; f1 is component towards point 1; f0 is component towards point 0;
            the trilinear coordinates sum to one and can be used as weights to interpolate z values at points
        """
        x = xy_array[..., 0].copy()
        y = xy_array[..., 1].copy()
        if self.origin is not None:
            x -= self.origin[0]
            y -= self.origin[1]
        jp = y / (self.t_side * root_3_by_2)
        j = np.floor(jp)
        mask = np.logical_or(j < 0, j >= self.nj - 1)
        fy = np.where(j % 2, (j + 1).astype(float) - jp, jp - j.astype(float))
        ip = x / self.t_side - 0.5 * fy
        i = np.floor(ip)
        mask = np.logical_or(mask, np.logical_or(i < 0, i >= self.ni - 1))
        fx = ip - i.astype(float)
        i *= 2
        am = np.where(fx > 1.0 - fy)
        i[am] += 1
        fx[am] -= 1.0 - fy[am]
        fy[am] = 1.0 - fy[am]
        j[mask] = -1
        i[mask] = -1
        fx[mask] = np.NaN
        fy[mask] = np.NaN

        return (np.stack((j, i), axis = -1), np.stack((1.0 - (fx + fy), fx, fy), axis = -1))

    def ji_and_weights_for_xy(self, xy):
        """Returns triplet of point ji indices and triplet of weights for interpolation at point xy.

        arguments:
            xy (pair of floats): coordinates x, y of a point of interest

        returns:
            numpy int array of shape (3, 2) and numpy triple float; int array holds mesh ji indices of three
            vertices of triangle containing xy; float triplet contains corresponding weights (summing to
            one) which can be used to interpolate z values at point xy
        """
        tji, tc = self.tji_tc_for_xy(xy)
        if tji is None:
            return (None, None)
        ji = self.tri_nodes_for_tji(tji)
        return (ji, tc)

    def ji_and_weights_for_xy_array(self, xy_array):
        """Returns arrays of triplet of point ji indices and triplet of weights for interpolation at points xy_array.

        arguments:
            xy_array (numpy float array of shape(..., 2 or 3)): coordinates x, y of points of interest

        returns:
            numpy int array of shape (..., 3, 2) and numpy float array of shape (..., 3); int array holds mesh ji indices
            of three vertices of triangles containing xy points; float triplets contain corresponding weights (summing to
            one per triangle) which can be used to interpolate z values at points xy_array
        """
        tji, tc = self.tji_tc_for_xy_array(xy_array)
        ji = self.tri_nodes_for_tji_array(tji)
        return (ji, tc)

    def interpolated_z(self, xy):
        """Returns z value interpolated trilinearly at point xy."""
        ji, tc = self.ji_and_weights_for_xy(xy)
        if ji is None:
            return None
        p = self.full_array_ref()
        z = np.empty(3, dtype = float)
        for vertex in range(3):
            z[vertex] = p[ji[vertex, 0], ji[vertex, 1], 2]
        return np.sum(tc * z)

    def interpolated_z_array(self, xy_array):
        """Returns z values interpolated trilinearly at points xy_array."""
        ji, tc = self.ji_and_weights_for_xy_array(xy_array)
        p = self.full_array_ref()
        z = np.empty(ji.shape[:-1], dtype = float)
        for vertex in range(3):
            # note: ji may have value -1 at some points; p should evaluate to arbitrary last point for that; tc will be NaN
            z[..., vertex] = p[ji[..., vertex, 0], ji[..., vertex, 1], 2]
        return np.sum(tc * z, axis = -1)

    def tri_nodes_for_tji(self, tji):
        """Return mesh node indices, shape (3, 2), for triangle tji (tj, ti)."""
        j, i = tji
        tn = np.zeros((3, 2), dtype = int)
        j_odd = j % 2
        i2, i_odd = divmod(i, 2)
        assert 0 <= j < self.nj - 1 and 0 <= i < 2 * (self.ni - 1)
        if i_odd:
            base_j = j + 1 - j_odd
            tip_j = j + j_odd
        else:
            base_j = j + j_odd
            tip_j = j + 1 - j_odd
        tn[0, 0] = base_j
        tn[1, 0] = base_j
        tn[2, 0] = tip_j
        tn[0, 1] = i2
        tn[1, 1] = i2 + 1
        tn[2, 1] = i2 + i_odd
        return tn

    def tri_nodes_for_tji_array(self, tji_array):
        """Return mesh node indices, shape (..., 3, 2), for triangles tji_array (..., 2) being (tj, ti)."""
        j = tji_array[..., 0]
        i = tji_array[..., 1]
        tn_shape = tuple(list(tji_array.shape[:-1]) + [3, 2])
        tn = np.zeros(tn_shape, dtype = int)
        j_odd = j % 2
        i2, i_odd = np.divmod(i, 2)
        mask = np.logical_or(np.logical_or(j < 0, j >= self.nj - 1), np.logical_or(i < 0, i >= 2 * (self.ni - 1)))
        base_j = np.where(i_odd, j + 1 - j_odd, j + j_odd)
        tip_j = np.where(i_odd, j + j_odd, j + 1 - j_odd)
        tn[..., 0, 0] = base_j
        tn[..., 1, 0] = base_j
        tn[..., 2, 0] = tip_j
        tn[..., 0, 1] = i2
        tn[..., 1, 1] = i2 + 1
        tn[..., 2, 1] = i2 + i_odd
        tn[mask] = -1
        return tn

    def all_tri_nodes(self):
        """Returns array of mesh node indices for all triangles, shape (nj - 1, 2 * (ni - 1), 3, 2)."""
        tna = np.zeros((self.nj - 1, 2 * (self.ni - 1), 3, 2), dtype = int)
        # set mesh j indices
        tna[:, :, 0, 0] = np.expand_dims(np.arange(self.nj - 1, dtype = int), axis = -1)
        tna[1::2, ::2, 0, 0] += 1
        tna[::2, 1::2, 0, 0] += 1
        tna[:, :, 1, 0] = tna[:, :, 0, 0]
        tna[:, :, 2, 0] = tna[:, :, 0, 0] + 1
        tna[1::2, ::2, 2, 0] -= 2
        tna[::2, 1::2, 2, 0] -= 2
        # set mesh i indices
        tna[:, ::2, 0, 1] = np.expand_dims(np.arange(self.ni - 1, dtype = int), axis = 0)
        tna[:, 1::2, 0, 1] = tna[:, ::2, 0, 1]
        tna[:, :, 1, 1] = tna[:, :, 0, 1] + 1
        tna[:, :, 2, 1] = tna[:, :, 0, 1]
        tna[:, 1::2, 2, 1] += 1
        return tna

    def triangles_and_points(self):
        """Returns node indices and xyz points in form suitable for a Surface (triangulated set)."""
        tna = self.all_tri_nodes()
        composite_ji = tna[:, :, :, 0] * self.ni + tna[:, :, :, 1]
        return (composite_ji.reshape((-1, 3)), self.full_array_ref().reshape((-1, 3)))

    def tji_for_triangle_index(self, ti):
        """Return triangle tji (tj, ti) for triangle index in Surface (triangulated set) protocol."""
        assert 0 <= ti < (self.nj - 1) * (self.ni - 1) * 2
        return divmod(ti, 2 * (self.ni - 1))

    def triangle_index_for_tji(self, tji):
        """Return triangle index in Surface (triangulated set) protocol for triangle tji (tj, ti)."""
        tj, ti = tji
        assert 0 <= tj < self.nj - 1 and 0 <= ti < 2 * (self.ni - 1)
        return 2 * (self.ni - 1) * tj + ti

    def tri_nodes_in_triangles(self, triangles):
        """Return indices of nodes of this tri mesh which are within other triangles, in xy space.

        arguments:
            triangles (numpy float array of shape (N, 3, 2)): other triangles' vertices in 2D (last axis is x,y)

        returns:
            numpy int array (list-like) of shape (M, 3) where last axis is (triangle index, nj, ni)

        note:
            other triangles must be in the same crs as this tri mesh; they do not need to be equilateral
        """

        assert triangles.ndim == 3 and triangles.shape[1] == 3 and triangles.shape[2] == 2

        tp = triangles.copy()
        if self.origin is not None:
            tp[:] -= np.expand_dims(np.expand_dims(self.origin[:2], axis = 0), axis = 0)

        # test odd node j rows of tri mesh points; note that vec function assumes a half 'cell' offset!
        tn_b = vec.points_in_triangles_aligned_optimised(self.ni, self.nj // 2, self.t_side,
                                                         2.0 * root_3_by_2 * self.t_side, tp)
        tn_b[:, 1] *= 2  # node j
        tn_b[:, 1] += 1

        # shift other triangles' points so as to compensate for half cell offset, for even j node rows
        offset = np.array((0.5, root_3_by_2), dtype = float) * self.t_side
        tp[:] += np.expand_dims(np.expand_dims(offset, axis = 0), axis = 0)

        # test even node j rows of tri mesh points
        tn_a = vec.points_in_triangles_aligned_optimised(self.ni, (self.nj + 1) // 2, self.t_side, root_3 * self.t_side,
                                                         tp)
        tn_a[:, 1] *= 2  # node j

        return np.concatenate((tn_a, tn_b), axis = 0)

    def edge_zero_crossings(self, z_values = None):
        """Returns numpy list of points from edges where z values cross zero (or given value).

        arguments:
            z_values (numpy float array of shape (nj, ni), optional): if present, these values
                are used to determine the crossing points, though returned xyz values are from
                the tri mesh's array

        returns:
            numpy float array of shape (N, 3) being points on tri mesh triangle edges where
            z values cross zero
        """

        xyz = self.full_array_ref()
        if z_values is None:
            z_values = xyz[..., 2]
        else:
            assert z_values.shape == (self.nj, self.ni)
        abs_ze = np.expand_dims(np.abs(z_values), axis = -1)
        p = np.empty((0, 3), dtype = float)

        # i edges
        crossing = np.where(z_values[:, :-1] * z_values[:, 1:] < 0.0)  # zero crossing i edge indices
        if len(crossing[0]):
            s = abs_ze[:, :-1, :] + abs_ze[:, 1:, :]
            # TODO: ignore divide by zero
            w = (xyz[:, :-1, :] * abs_ze[:, 1:, :] + xyz[:, 1:, :] * abs_ze[:, :-1, :]) / s
            # TODO: unignore, and use midpoint of those edges?
            p = np.concatenate((p, w[crossing]))

        # one set of j edges
        crossing = np.where(z_values[:-1, :] * z_values[1:, :] < 0.0)  # zero crossing j edge indices
        if len(crossing[0]):
            s = abs_ze[:-1, :, :] + abs_ze[1:, :, :]
            w = (xyz[:-1, :, :] * abs_ze[1:, :, :] + xyz[1:, :, :] * abs_ze[:-1, :, :]) / s
            p = np.concatenate((p, w[crossing]))

        # other set of j edges (even j base)
        crossing = np.where(z_values[:-1:2, 1:] * z_values[1::2, :-1] < 0.0)  # zero crossing j edge indices
        if len(crossing[0]):
            s = abs_ze[:-1:2, 1:, :] + abs_ze[1::2, :-1, :]
            w = (xyz[:-1:2, 1:, :] * abs_ze[1::2, :-1, :] + xyz[1::2, :-1, :] * abs_ze[:-1:2, 1:, :]) / s
            p = np.concatenate((p, w[crossing]))

        # other set of j edges (odd j base)
        crossing = np.where(z_values[1:-1:2, :-1] * z_values[2::2, 1:] < 0.0)  # zero crossing j edge indices
        if len(crossing[0]):
            s = abs_ze[1:-1:2, :-1, :] + abs_ze[2::2, 1:, :]
            w = (xyz[1:-1:2, :-1, :] * abs_ze[2::2, 1:, :] + xyz[2::2, 1:, :] * abs_ze[1:-1:2, :-1, :]) / s
            p = np.concatenate((p, w[crossing]))

        return p

    def axial_edge_crossings(self, axis, value = 0.0):
        """Returns list-like array of points interpolated from edges that cross value in axis.

        arguments:
            axis (int): xyz axis for crossings; 0: x, 1: y, 2: z
            value (float, default 0.0): the value to detect crossings for

        returns:
            numpy float array of shape (N, 3) being the points on tri mesh edges at which the axial crossing
            occurs
        """

        xyz = self.full_array_ref()
        zero = xyz[..., axis] - value
        abs_zero = np.expand_dims(np.abs(zero), axis = -1)
        p = np.empty((0, 3), dtype = float)

        if axis == 0:  # x axis

            xi, xr = divmod(value, self.t_side)
            if xi >= 0.0 and xi + xr < (float(self.ni) - 0.5) * self.t_side:
                i = round(xi)
                xrf = xr / self.t_side
                if xi + xr < float(self.ni - 1) * self.t_side:
                    # I edges, even node j
                    p = xyz[::2, i] * (1.0 - xrf) + xyz[::2, i + 1] * xrf
                if i or xrf > 0.5:
                    vp = value - 0.5 * self.t_side
                    xip, xrp = divmod(vp, self.t_side)
                    xrp /= self.t_side
                    # i edges, odd node j
                    ip = round(xip)
                    pp = xyz[1::2, ip] * (1.0 - xrp) + xyz[1::2, ip + 1] * xrp
                    p = np.concatenate((p, pp))
                if xr != 0.0 and xrf != 0.5:
                    # j edges of one sort ot the other
                    if xrf < 0.5:
                        xrp = 2.0 * xrf
                        pp = xyz[:-1:2, i] * (1.0 - xrp) + xyz[1::2, i] * xrp
                        p = np.concatenate((p, pp))
                        pp = xyz[1:-1:2, i] * xrp + xyz[2::2, i] * (1.0 - xrp)
                        p = np.concatenate((p, pp))
                    else:
                        xrp = 2.0 * (1.0 - xrf)
                        pp = xyz[:-1:2, i + 1] * (1.0 - xrp) + xyz[1::2, i] * xrp
                        p = np.concatenate((p, pp))
                        pp = xyz[1:-1:2, i] * xrp + xyz[2::2, i + 1] * (1.0 - xrp)
                        p = np.concatenate((p, pp))

        elif axis == 1:

            yi, yr = divmod(value, self.t_side * root_3_by_2)

            if yi >= 0 and yi + yr <= (self.nj - 1) * self.t_side * root_3_by_2:
                j = round(yi)
                if yr == 0.0:  # add first and last points from in-line edges
                    p = np.array([xyz[j, 0], xyz[j, -1]], dtype = float)
                else:
                    yrp = yr / (self.t_side * root_3_by_2)
                    p = xyz[j, :] * (1.0 - yrp) + xyz[j + 1, :] * yrp
                    if j % 2:
                        pp = xyz[j, 1:] * (1.0 - yrp) + xyz[j + 1, :-1] * yrp
                    else:
                        pp = xyz[j, :-1] * (1.0 - yrp) + xyz[j + 1, 1:] * yrp
                p = np.concatenate((p, pp))

        else:  # z axis

            # i edges
            crossing = np.where(zero[:, :-1] * zero[:, 1:] < 0.0)  # zero crossing i edge indices
            if len(crossing[0]):
                s = abs_zero[:, :-1] + abs_zero[:, 1:]
                # TODO: ignore divide by zero
                w = (xyz[:, :-1, :] * abs_zero[:, 1:] + xyz[:, 1:, :] * abs_zero[:, :-1]) / s
                # TODO: unignore, and use midpoint of those edges?
                p = np.concatenate((p, w[crossing]))

            # one set of j edges
            crossing = np.where(zero[:-1, :] * zero[1:, :] < 0.0)  # zero crossing j edge indices
            if len(crossing[0]):
                s = abs_zero[:-1, :] + abs_zero[1:, :]
                w = (xyz[:-1, :, :] * abs_zero[1:, :] + xyz[1:, :, :] * abs_zero[:-1, :]) / s
                p = np.concatenate((p, w[crossing]))

            # other set of j edges (even j base)
            crossing = np.where(zero[:-1:2, 1:] * zero[1::2, :-1] < 0.0)  # zero crossing j edge indices
            if len(crossing[0]):
                s = abs_zero[:-1:2, 1:] + abs_zero[1::2, :-1]
                w = (xyz[:-1:2, 1:, :] * abs_zero[1::2, :-1] + xyz[1::2, :-1, :] * abs_zero[:-1:2, 1:]) / s
                p = np.concatenate((p, w[crossing]))

            # other set of j edges (odd j base)
            crossing = np.where(zero[1:-1:2, :-1] * zero[2::2, 1:] < 0.0)  # zero crossing j edge indices
            if len(crossing[0]):
                s = abs_zero[1:-1:2, :-1] + abs_zero[2::2, 1:]
                w = (xyz[1:-1:2, :-1, :] * abs_zero[2::2, 1:] + xyz[2::2, 1:, :] * abs_zero[1:-1:2, :-1]) / s
                p = np.concatenate((p, w[crossing]))

        return p



"""Multiprocessing module for running specific functions concurrently."""

__all__ = [
    'function_multiprocessing', 'find_faces_to_represent_surface_regular_wrapper',
    'mesh_from_regular_grid_column_property_wrapper', 'mesh_from_regular_grid_column_property_batch',
    'blocked_well_from_trajectory_wrapper', 'blocked_well_from_trajectory_batch'
]

from ._multiprocessing import function_multiprocessing
from .wrappers.grid_surface_mp import find_faces_to_represent_surface_regular_wrapper
from .wrappers.mesh_mp import mesh_from_regular_grid_column_property_wrapper, mesh_from_regular_grid_column_property_batch
from .wrappers.blocked_well_mp import blocked_well_from_trajectory_wrapper, blocked_well_from_trajectory_batch

# Set "module" attribute of all public objects to this path.
for _name in __all__:
    _obj = eval(_name)
    if hasattr(_obj, "__module__"):
        _obj.__module__ = __name__
